import logging
import scipy
import collections
import numpy as np
from math import pi, sqrt, exp
from scipy import stats, interpolate, signal
from scipy.signal import convolve, gaussian


def getGumbel(dblE,dblV,dblX):
    """"Calculate p-value and z-score for maximum value of N samples drawn from Gaussian
       [dblP,dblZ] = getGumbel(dblE,dblV,dblX)

        input:
        - dblE: mean of distribution of maximum values
        - dblV: variance of distribution of maximum values
        - dblX: maximum value to express in quantiles of Gumbel

        output:
        - dblP; p-value for dblX (chance that sample originates from distribution given by dblE/dblV)
        - dblZ; z-score corresponding to P

    Version history:
    1.0 - June 17, 2020, Created by Jorrit Montijn translated by Alexander Heimel

    Sources:
    Baglivo (2005), ISBN: 9780898715668
    Elfving (1947), https://doi.org/10.1093/biomet/34.1-2.111
    Royston (1982), DOI: 10.2307/2347982
    https://stats.stackexchange.com/questions/394960/variance-of-normal-order-statistics
    https://stats.stackexchange.com/questions/9001/approximate-order-statistics-for-normal-random-variables
    https://en.wikipedia.org/wiki/Extreme_value_theory
    https://en.wikipedia.org/wiki/Gumbel_distribution
    """

    ## define Gumbel parameters from mean and variance
    #derive beta parameter from variance
    dblBeta = sqrt(6) * sqrt(dblV) / pi

    # define Euler-Mascheroni constant
    dblEulerMascheroni = 0.5772156649015328606065120900824 #vpa(eulergamma)

    # derive mode from mean, beta and E-M constant
    dblMode = dblE - dblBeta * dblEulerMascheroni

    # define Gumbel cdf
    ###    fGumbelCDF = @(x) exp(-exp(-((x(:)-dblMode)./dblBeta)));
    fGumbelCDF = lambda x : exp(-exp(-((x-dblMode) /dblBeta)))

    ## calculate output variables
    # calculate cum dens at X
    dblGumbelCDF = fGumbelCDF(dblX)
    # define p-value
    dblP = 1-dblGumbelCDF
    # transform to output z-score
    ### dblZ = -norminv(dblP/2);
    dblZ = -scipy.stats.norm.ppf(dblP/2)

    # approximation for large X
    ### dblP[isinf(dblZ)] = exp( (dblMode-dblX(isinf(dblZ)))./dblBeta ) ;
    if dblZ>1E30:
        dblP = exp( (dblMode-dblX) / dblBeta )
    # transform to output z-score
    ### dblZ = -norminv(dblP/2);
    dblZ = -scipy.stats.norm.ppf(dblP/2)

    return dblP,dblZ

def flatten(l):
    for el in l:
        if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el

def getTempOffset(vecSpikeT, vecSpikeTimes, vecStimUseOnTime, dblUseMaxDur):
    """Calculate temporal offset vectors across folds and offsets.

    Syntax:
    [vecThisDiff,vecThisFrac,vecThisFracLinear] =
        getTempOffset(vecSpikeT,vecSpikeTimes,vecStimUseOnTime,dblUseMaxDur)
    """

    # go through trials to build spike time vector
    cellSpikeTimesPerTrial = []
    for intEvent, dblStartT in enumerate(vecStimUseOnTime):
        # get times
        dblStopT = dblStartT + dblUseMaxDur

        # build trial assignment
        cellSpikeTimesPerTrial.append(vecSpikeTimes[(vecSpikeTimes < dblStopT)
                                                    & (vecSpikeTimes > dblStartT)] - dblStartT)

    # get spikes in fold
    vecThisSpikeT = np.concatenate(cellSpikeTimesPerTrial)

    # get real fractions for training set
    vecThisSpikeTimes = np.sort(np.concatenate(([0], vecThisSpikeT, [dblUseMaxDur])))
    vecThisSpikeFracs = np.linspace(0, 1, vecThisSpikeTimes.shape[0])
    vecThisFrac = interpolate.interp1d(vecThisSpikeTimes, vecThisSpikeFracs)(vecSpikeT)

    # get linear fractions
    vecThisFracLinear = vecSpikeT / dblUseMaxDur

    # calc difference
    vecThisDiff = vecThisFrac - vecThisFracLinear

    return vecThisDiff, vecThisFrac, vecThisFracLinear


def getPeak(vecData, vecT, vecRestrictRange=(-np.inf,np.inf), intSwitchZ=1):
    """Returns highest peak time, width, and location. Syntax:
        [dblPeakValue,dblPeakTime,dblPeakWidth,vecPeakStartStop,intPeakLoc,vecPeakStartStopIdx] = getPeak(vecData,vecT,vecRestrictRange)

    Required input:
        - vecData [N x 1]: values

    Optional inputs:
        - vecT [N x 1]: timestamps corresponding to vecData (default: [1:N])
        - vecRestrictRange: restrict peak to lie within vecRestrictRange(1) and vecRestrictRange(end)

    Outputs:
        - dblPeakTime: time of peak
        - dblPeakWidth: width of peak
        - vecPeakStartStop: start/stop times of peak
        - intPeakLoc: index of peak
        - vecPeakStartStopIdx: start/stop indices of peak

    Version history:
    1.0 - June 19, 2020, Created by Jorrit Montijn, Translated to python by Alexander Heimel
    """

    # check inputs
    if len(vecT) == 0:
        vecT = np.arange(len(vecData))

    # z-score
    if intSwitchZ == 1:
        vecDataZ = stats.zscore(vecData)
    elif intSwitchZ == 2:
        dblMu = np.mean(vecData[vecT<0.02])
        vecDataZ = (vecData - dblMu) / np.std(vecData)
    else:
        vecDataZ = vecData

    # get most prominent POSITIVE peak times
    vecLocsPos, peakProps = signal.find_peaks(vecDataZ, threshold=0, prominence=-np.inf)
    vecValsPos = vecDataZ[vecLocsPos]
    vecPromsPos = peakProps['prominences']

    # remove peaks outside window
    indKeepPeaks = (vecT[vecLocsPos] >= vecRestrictRange[0]) & (vecT[vecLocsPos] <= vecRestrictRange[1])

    if np.sum(indKeepPeaks) == 0:
        dblMaxPosVal = None
    else:
        # select peak
        vecValsPos = vecValsPos[indKeepPeaks]
        vecLocsPos = vecLocsPos[indKeepPeaks]
        vecPromsPos = vecPromsPos[indKeepPeaks]
        intPosIdx = np.argmax(vecValsPos)
        dblMaxPosVal = vecValsPos[intPosIdx]

    # get most prominent NEGATIVE peak times
    vecLocsNeg, peakProps = signal.find_peaks(-vecDataZ, threshold=0, prominence=-np.inf)
    vecValsNeg = -vecDataZ[vecLocsNeg]
    vecPromsNeg = peakProps['prominences']

    # remove peaks outside window
    indKeepPeaks = (vecT[vecLocsNeg] >= vecRestrictRange[0]) & (vecT[vecLocsNeg] <= vecRestrictRange[1])

    if np.sum(indKeepPeaks) == 0:
        dblMaxNegVal = None
    else:
        # select peak
        vecValsNeg = vecValsNeg[indKeepPeaks]
        vecLocsNeg = vecLocsNeg[indKeepPeaks]
        vecPromsNeg = vecPromsNeg[indKeepPeaks]
        intNegIdx = np.argmax(vecValsNeg)
        dblMaxNegVal = vecValsNeg[intNegIdx]

    if dblMaxPosVal is None and dblMaxNegVal is None :
        indPeakMembers = None
    elif ((dblMaxPosVal is not None and dblMaxNegVal is None)
          or (dblMaxPosVal is not None and (np.abs(dblMaxPosVal) >= np.abs(dblMaxNegVal)))):
        intIdx = intPosIdx
        intPeakLoc = vecLocsPos[intIdx]
        dblPeakProm = vecPromsPos[intIdx]
        dblCutOff = vecDataZ[intPeakLoc] - dblPeakProm / 2
        indPeakMembers = (vecDataZ > dblCutOff)
    elif ((dblMaxPosVal is None and dblMaxNegVal is not None)
          or (dblMaxNegVal is not None and (np.abs(dblMaxPosVal) < np.abs(dblMaxNegVal)))):
        intIdx = intNegIdx
        intPeakLoc = vecLocsNeg[intIdx]
        dblPeakProm = vecPromsNeg[intIdx]
        dblCutOff = vecDataZ[intPeakLoc] + dblPeakProm / 2
        indPeakMembers = (vecDataZ < dblCutOff)

    if indPeakMembers is not None:
        # get potential starts/stops
        ### vecPeakStarts = find(diff(indPeakMembers)==1);
        vecPeakStarts = np.where(np.diff([float(f) for f in indPeakMembers])==1)[0]
        ### vecPeakStops = find(diff(indPeakMembers)==-1);
        vecPeakStops = np.where(np.diff([float(f) for f in indPeakMembers])==-1)[0]
        if indPeakMembers[0] == True:
            ### vecPeakStarts = [1 vecPeakStarts(:)'];
            vecPeakStarts = [0] + vecPeakStarts
        ### if indPeakMembers(end) == 1
        if indPeakMembers[-1] == True:
            ### vecPeakStops = [vecPeakStops(:)' numel(indPeakMembers)];
            vecPeakStops = vecPeakStops + [len(indPeakMembers)-1]

        # find closest points
        ###    intPeakStart = intPeakLoc-min(intPeakLoc - vecPeakStarts(vecPeakStarts<intPeakLoc));

        intPeakStart = intPeakLoc - np.min(intPeakLoc - vecPeakStarts[vecPeakStarts<intPeakLoc])
        intPeakStop = intPeakLoc + np.min(vecPeakStops[vecPeakStops>=intPeakLoc] - intPeakLoc)
        dblPeakStartT = vecT[intPeakStart]
        if intPeakStop >= vecT.shape[0]:
            intPeakStop = vecT.shape[0] - 1
        dblPeakStopT = vecT[intPeakStop]
        # assign peak data
        dblPeakValue = vecData[intPeakLoc]
        dblPeakTime = vecT[intPeakLoc]
        dblPeakWidth = dblPeakStopT - dblPeakStartT
        vecPeakStartStop = [dblPeakStartT, dblPeakStopT]
        vecPeakStartStopIdx = [intPeakStart, intPeakStop]
    else:
        # assign placeholder peak data
        dblPeakValue = np.nan
        dblPeakTime = np.nan
        dblPeakWidth = np.nan
        vecPeakStartStop = [np.nan, np.nan]
        intPeakLoc = None
        vecPeakStartStopIdx = [None, None]

    return dblPeakValue, dblPeakTime, dblPeakWidth, vecPeakStartStop, intPeakLoc, vecPeakStartStopIdx


def getOnset(vecData,vecT,dblPeakT,vecRestrictRange,intSwitchZ=1):
    """Returns peak onset. Syntax:
        [dblOnset,dblValue] = getOnset(vecData,vecT,dblPeakT,vecRestrictRange)

    Required input:
        - vecData [N x 1]: values

    Optional inputs:
        - vecT [N x 1]: timestamps corresponding to vecData (default: [1:N])
        - dblPeakT (float): timestamp corresponding to peak
        - vecRestrictRange [2 x 1]: restrict peak to lie within vecRestrictRange(1) and vecRestrictRange(end)

    Outputs:
        - dblOnset: time of peak onset (first crossing half-height of peak)
        - dblValue: value at peak onset

    Version history:
    1.0 - June 19, 2020 Created by Jorrit Montijn, Translated to Python by Alexander Heimel
    """

    ##
    # check inputs
    if vecT == []:
        vecT = np.arange(len(vecData))

    if vecRestrictRange == None:
        ### vecRestrictRange = [min(vecT) min(vecT)+range(vecT)/2];
        vecRestrictRange = (np.min(vecT), np.min(vecT) + (np.max(vecT)-np.min(vecT))/2)

    # z-score
    if intSwitchZ == 1:
        vecDataZ = stats.zscore(vecData)
    elif intSwitchZ == 2:
        dblMu = np.mean(vecData[vecT<0.02])
        vecDataZ = (vecData - dblMu) / np.std(vecData)
    else:
        vecDataZ = vecData

    if dblPeakT==None:
        dblPeakT = getPeak(vecDataZ, vecT, vecRestrictRange, 0)[1]

    # remove time points outside restricted range
    indRemove = (vecT < vecRestrictRange[0]) | (vecT > vecRestrictRange[1])
    vecCropT = vecT[np.invert(indRemove)]
    vecDataZ = vecDataZ[np.invert(indRemove)]

    # calculate first timepoint crossing half-height of peak
    intPeakIdx = np.argmin(abs(vecCropT-dblPeakT))

    dblPeakVal = vecDataZ[intPeakIdx]
    dblBaseVal = vecDataZ[0]
    dblThresh = (dblPeakVal - dblBaseVal)/2 + dblBaseVal
    ### if dblThresh > 0
    ###     intOnsetIdx = find(vecDataZ >= dblThresh,1,'first');
    ### else
    ###     intOnsetIdx = find(vecDataZ <= dblThresh,1,'first');
    ### end
    if dblThresh > 0:
        intOnsetIdx = np.where(vecDataZ >= dblThresh)[0]
    else:
        intOnsetIdx = np.where(vecDataZ <= dblThresh)[0]

    if len(intOnsetIdx)>0:
        intOnsetIdx = intOnsetIdx[0]
        dblOnset = vecCropT[intOnsetIdx]
        dblValue = vecData[vecT > dblOnset]
        dblValue = dblValue[0]
    else:
        dblOnset = None
        dblValue = None

    return dblOnset, dblValue, dblBaseVal, dblPeakT


def calculatePeths(
        spike_times, spike_clusters, cluster_ids, align_times, pre_time=0.2,
        post_time=0.5, bin_size=0.025, smoothing=0.025, return_fr=True):
    """
    Calcluate peri-event time histograms; return means and standard deviations
    for each time point across specified clusters
    
    Code modified from Brainbox library of the International Brain Laboratory
    https://github.com/int-brain-lab/ibllib/blob/master/brainbox/singlecell.py
    
    :param spike_times: spike times (in seconds)
    :type spike_times: array-like
    :param spike_clusters: cluster ids corresponding to each event in `spikes`
    :type spike_clusters: array-like
    :param cluster_ids: subset of cluster ids for calculating peths
    :type cluster_ids: array-like
    :param align_times: times (in seconds) to align peths to
    :type align_times: array-like
    :param pre_time: time (in seconds) to precede align times in peth
    :type pre_time: float
    :param post_time: time (in seconds) to follow align times in peth
    :type post_time: float
    :param bin_size: width of time windows (in seconds) to bin spikes
    :type bin_size: float
    :param smoothing: standard deviation (in seconds) of Gaussian kernel for
        smoothing peths; use `smoothing=0` to skip smoothing
    :type smoothing: float
    :param return_fr: `True` to return (estimated) firing rate, `False` to return spike counts
    :type return_fr: bool
    :return: peths, binned_spikes
    :rtype: peths: Bunch({'mean': peth_means, 'std': peth_stds, 'tscale': ts, 'cscale': ids})
    :rtype: binned_spikes: np.array (n_align_times, n_clusters, n_bins)
    """

    # initialize containers
    n_offset = 5 * int(np.ceil(smoothing / bin_size))  # get rid of boundary effects for smoothing
    n_bins_pre = int(np.ceil(pre_time / bin_size)) + n_offset
    n_bins_post = int(np.ceil(post_time / bin_size)) + n_offset
    n_bins = n_bins_pre + n_bins_post
    binned_spikes = np.zeros(shape=(len(align_times), len(cluster_ids), n_bins))

    # build gaussian kernel if requested
    if smoothing > 0:
        w = n_bins - 1 if n_bins % 2 == 0 else n_bins
        window = gaussian(w, std=smoothing / bin_size)
        # half (causal) gaussian filter
        # window[int(np.ceil(w/2)):] = 0
        window /= np.sum(window)
        binned_spikes_conv = np.copy(binned_spikes)

    ids = np.unique(cluster_ids)

    # filter spikes outside of the loop
    idxs = np.bitwise_and(spike_times >= np.min(align_times) - (n_bins_pre + 1) * bin_size,
                          spike_times <= np.max(align_times) + (n_bins_post + 1) * bin_size)
    idxs = np.bitwise_and(idxs, np.isin(spike_clusters, cluster_ids))
    spike_times = spike_times[idxs]
    spike_clusters = spike_clusters[idxs]

    # compute floating tscale
    tscale = np.arange(-n_bins_pre, n_bins_post + 1) * bin_size
    # bin spikes
    for i, t_0 in enumerate(align_times):
        # define bin edges
        ts = tscale + t_0
        # filter spikes
        idxs = np.bitwise_and(spike_times >= ts[0], spike_times <= ts[-1])
        i_spikes = spike_times[idxs]
        i_clusters = spike_clusters[idxs]

        # bin spikes similar to bincount2D: x = spike times, y = spike clusters
        xscale = ts
        xind = (np.floor((i_spikes - np.min(ts)) / bin_size)).astype(np.int64)
        yscale, yind = np.unique(i_clusters, return_inverse=True)
        nx, ny = [xscale.size, yscale.size]
        ind2d = np.ravel_multi_index(np.c_[yind, xind].transpose(), dims=(ny, nx))
        r = np.bincount(ind2d, minlength=nx * ny, weights=None).reshape(ny, nx)

        # store (ts represent bin edges, so there are one fewer bins)
        bs_idxs = np.isin(ids, yscale)
        binned_spikes[i, bs_idxs, :] = r[:, :-1]

        # smooth
        if smoothing > 0:
            idxs = np.where(bs_idxs)[0]
            for j in range(r.shape[0]):
                binned_spikes_conv[i, idxs[j], :] = convolve(
                    r[j, :], window, mode='same', method='auto')[:-1]

    # average
    if smoothing > 0:
        binned_spikes_ = np.copy(binned_spikes_conv)
    else:
        binned_spikes_ = np.copy(binned_spikes)
    if return_fr:
        binned_spikes_ /= bin_size

    peth_means = np.mean(binned_spikes_, axis=0)
    peth_stds = np.std(binned_spikes_, axis=0)
    peth_sems = np.std(binned_spikes_, axis=0) / np.sqrt(align_times.shape[0])

    if smoothing > 0:
        peth_means = peth_means[:, n_offset:-n_offset]
        peth_stds = peth_stds[:, n_offset:-n_offset]
        binned_spikes = binned_spikes[:, :, n_offset:-n_offset]
        tscale = tscale[n_offset:-n_offset]

    # package output
    tscale = (tscale[:-1] + tscale[1:]) / 2
    peths = dict({'means': peth_means, 'stds': peth_stds, 'sems': peth_sems,
                  'tscale': tscale, 'cscale': ids})
    return peths, binned_spikes


# -*- coding: utf-8 -*-
import numpy as np
import logging

# from zetapy import msd
from scipy import stats
from zetapy.dependencies import (calcZetaOne, calcZetaTwo, getTempOffsetOne, flatten, findfirst)
from zetapy.ifr_dependencies import (getMultiScaleDeriv, getPeak, getOnset)
from zetapy.plot_dependencies import calculatePeths, plotzeta, plotzeta2, plottszeta, plottszeta2
from zetapy.ts_dependencies import calcTsZetaOne, calcTsZetaTwo, getPseudoTimeSeries, getTsRefT, getInterpolatedTimeSeries

# %% two-sample time-series zeta test


def zetatstest2(vecTime1, vecValue1, arrEventTimes1, vecTime2, vecValue2, arrEventTimes2,
                dblUseMaxDur=None, intResampNum=250, boolPlot=False, boolDirectQuantile=False, dblSuperResFactor=100):
    """
    Calculates two-sample time-series zeta-test

    Heimel, J.A., Meijer, G.T., Montijn, J.S. (2023). A new family of statistical tests for responses
    in point-event and time-series data for one- and two-sample comparisons. bioRxiv

    Montijn, J.S., Seignette, K., Howlett, M.H., Cazemier, J.L., Kamermans, M., Levelt, C.N.,
    and Heimel, J.A. (2021). A parameter-free statistical test for neuronal responsiveness.
    eLife 10, e71969.

    Syntax:
    dblZeta2P,dZETA2 = zetatest2(vecTime1,vecValue1,arrEventTimes1,vecTime2,vecValue2,arrEventTimes2,
                                 dblUseMaxDur=None,intResampNum=250,boolPlot=False,boolDirectQuantile=False,dblSuperResFactor=100)


    Parameters
    ----------
    vecTime1 [N x 1]: 1D array (float)
        timestamps in seconds corresponding to entries in vecValue1.
    vecValue1 [N x 1] : 1D array (float)
        values for condition 1 (e.g., dF/F0 activity).
    arrEventTimes1 : 1D or 2D array (float)
        event on times (s), or [T x 2] including event off times for condition 1 to calculate mean-rate difference.
    vecTime2 [N x 1]: 1D array (float)
        timestamps in seconds corresponding to entries in vecValue2.
    vecValue2 [N x 1] : 1D array (float)
        values for condition 2 (e.g., dF/F0 activity).
    arrEventTimes2 : 1D or 2D array (float)
        event on times (s), or [T x 2] including event off times for condition 2 to calculate mean-rate difference.

    Optional Parameters
    ----------
    dblUseMaxDur : float
        window length for calculating ZETA: ignore all entries beyond this duration after event onset
        (default: minimum of event onset to event onset)
    intResampNum : integer
        number of resamplings (default: 250)
        [Note: if your p-value is close to significance, you should increase this number to enhance the precision]
    boolPlot : boolean switch
        plotting switch (False: no plot, True: plot figure) (default: False)
    boolDirectQuantile: boolean
        switch to use the empirical null-distribution rather than the Gumbel approximation (default: False)
        [Note: requires many resamplings!]
    dblSuperResFactor; scalar
        upsampling of data when calculating zeta (default: 100)

    Returns
    -------
    dblZetaP : float
        p-value based on Zenith of Event-based Time-locked Anomalies for two-sample comparison
    dZETA : dict
        additional information of ZETA test
            dblZetaP; p-value based on Zenith of Event-based Time-locked Anomalies (same as above)
            dblZETA; responsiveness z-score (i.e., >2 is significant)
            dblMeanZ; z-score for mean-rate stim/base difference (i.e., >2 is significant)
            dblMeanP; p-value based on mean-rate stim/base difference
            dblZETADeviation; temporal deviation value underlying ZETA
            dblZetaT; time corresponding to ZETA
            intZetaIdx; entry corresponding to ZETA
            vecMu1; average spiking rate values per event underlying t-test for condition 1
            vecMu2; average spiking rate values per event underlying t-test for condition 2
            vecRefTime: timestamps of trace entries (corresponding to vecRealDiff/matRandDiff)
            vecRealDiff; difference between condition 1 and 2
            matRandDiff; random differences in cumulative density of spikes
            vecRealFrac1; cumulative spike vector of condition 1
            vecRealFrac2; cumulative spike vector of condition 2
            dblZETADeviation_InvSign; largest deviation of inverse sign to ZETA (i.e., -ZETA)
            dblZETATime_InvSign; time corresponding to -ZETA
            intZETAIdx_InvSign; entry corresponding to -ZETA
            dblUseMaxDur; window length used to calculate ZETA

    Version history:
    1.0 - 30 Oct 2023 Translated to python [by Jorrit Montijn]

    """

    # %% build placeholder outputs
    dblZetaP = 1.0
    dZETA = dict()

    # fill dZETA
    # ZETA significance
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = None
    # mean-rate significance
    dZETA['dblMeanZ'] = None
    dZETA['dblMeanP'] = None
    # data on ZETA peak
    dZETA['dblZETADeviation'] = None
    dZETA['dblZETATime'] = None
    dZETA['intZETAIdx'] = None
    # data underlying mean-rate test
    dZETA['vecMu1'] = None
    dZETA['vecMu2'] = None
    # inverse-sign ZETA
    dZETA['dblZETADeviation_InvSign'] = None
    dZETA['dblZETATime_InvSign'] = None
    dZETA['intZETAIdx_InvSign'] = None

    # derived from calcTsZetaTwo
    dZETA['vecRefTime'] = None
    dZETA['vecRealDiff'] = None
    dZETA['matRandDiff'] = None
    dZETA['vecRealFrac1'] = None
    dZETA['vecRealFrac2'] = None
    dZETA['matTracePerTrial1'] = None
    dZETA['matTracePerTrial2'] = None

    # dZETA['dblZetaP'] = None #<-updates automatically
    # dZETA['dblZETA'] = None #<-updates automatically
    # dZETA['intZETAIdx'] = None #<-updates automatically

    # window used for analysis
    dZETA['dblUseMaxDur'] = None

    # %% prep data and assert inputs are correct

    # vecTime1 and vecValue1 must be [N by 1] arrays
    assert len(vecTime1.shape) == len(
        vecValue1.shape) and vecTime1.shape == vecValue1.shape, "vecTime1 and vecValue1 have different shapes"
    assert (len(vecTime1.shape) == 1 or vecTime1.shape[1] == 1) and issubclass(
        vecTime1.dtype.type, np.floating), "Input vecTime1 is not a 1D float np.array with >2 spike times"
    vecTime1 = vecTime1.flatten()
    vecValue1 = vecValue1.flatten()
    vecReorder = np.argsort(vecTime1, axis=0)
    vecTime1 = vecTime1[vecReorder]
    vecValue1 = vecValue1[vecReorder]

    # vecTime2 and vecValue2 must be [N by 1] arrays
    assert len(vecTime2.shape) == len(
        vecValue2.shape) and vecTime2.shape == vecValue2.shape, "vecTime2 and vecValue2 have different shapes"
    assert (len(vecTime2.shape) == 1 or vecTime2.shape[1] == 1) and issubclass(
        vecTime2.dtype.type, np.floating), "Input vecTime2 is not a 1D float np.array with >2 spike times"
    vecTime2 = vecTime2.flatten()
    vecValue2 = vecValue2.flatten()
    vecReorder = np.argsort(vecTime2, axis=0)
    vecTime2 = vecTime2[vecReorder]
    vecValue2 = vecValue2[vecReorder]

    # ensure orientation and assert that arrEventTimes1 is a 1D or N-by-2 array of floats
    assert len(arrEventTimes1.shape) < 3 and issubclass(
        arrEventTimes1.dtype.type, np.floating), "Input arrEventTimes1 is not a 1D or 2D float np.array"
    if len(arrEventTimes1.shape) > 1:
        if arrEventTimes1.shape[1] < 3:
            pass
        elif arrEventTimes1.shape[0] < 3:
            arrEventTimes1 = arrEventTimes1.T
        else:
            raise Exception(
                "Input error: arrEventTimes1 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes1 = np.reshape(arrEventTimes1, (-1, 1))
    # define event starts
    vecEventStarts1 = arrEventTimes1[:, 0]

    # ensure orientation and assert that arrEventTimes2 is a 1D or N-by-2 array of floats
    assert len(arrEventTimes2.shape) < 3 and issubclass(
        arrEventTimes2.dtype.type, np.floating), "Input arrEventTimes2 is not a 1D or 2D float np.array"
    if len(arrEventTimes2.shape) > 1:
        if arrEventTimes2.shape[1] < 3:
            pass
        elif arrEventTimes2.shape[0] < 3:
            arrEventTimes2 = arrEventTimes2.T
        else:
            raise Exception(
                "Input error: arrEventTimes2 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes2 = np.reshape(arrEventTimes2, (-1, 1))
    # define event starts
    vecEventStarts2 = arrEventTimes2[:, 0]

    # check if number of events and values is sufficient
    if vecTime1.size < 3 or vecEventStarts1.size < 3:
        if vecTime1.size < 3:
            strMsg1 = f"Number of entries in time-series ({vecTime1.size}) is too few to calculate zeta; "
        else:
            strMsg1 = ""
        if vecEventStarts1.size < 3:
            strMsg2 = f"Number of events ({vecEventStarts1.size}) is too few to calculate zeta; "
        else:
            strMsg2 = ""

        logging.warning("zetatstest2: " + strMsg1 + strMsg2 + "defaulting to p=1.0")
        return dblZetaP, dZETA

    # check if number of events and values is sufficient
    if vecTime2.size < 3 or vecEventStarts2.size < 3:
        if vecTime2.size < 3:
            strMsg1 = f"Number of entries in time-series ({vecTime2.size}) is too few to calculate zeta; "
        else:
            strMsg1 = ""
        if vecEventStarts2.size < 3:
            strMsg2 = f"Number of events ({vecEventStarts2.size}) is too few to calculate zeta; "
        else:
            strMsg2 = ""

        logging.warning("zetatstest2: " + strMsg1 + strMsg2 + "defaulting to p=1.0")
        return dblZetaP, dZETA

    # is stop supplied?
    if len(arrEventTimes1.shape) > 1 and arrEventTimes1.shape[1] > 1 and len(arrEventTimes2.shape) > 1 and arrEventTimes2.shape[1] > 1:
        boolStopSupplied = True
        vecEventStops1 = arrEventTimes1[:, 1]
        vecEventOnDur1 = arrEventTimes1[:, 1] - arrEventTimes1[:, 0]
        vecEventStops2 = arrEventTimes2[:, 1]
        vecEventOnDur2 = arrEventTimes2[:, 1] - arrEventTimes2[:, 0]
        assert np.all(vecEventOnDur1 > 0) and np.all(vecEventOnDur2 >
                                                     0), "at least one event in arrEventTimes has a non-positive duration"
        vecMu1 = np.zeros(vecEventStops1.shape)
        vecMu2 = np.zeros(vecEventStops2.shape)

    else:
        boolStopSupplied = False
        dblMeanZ = np.nan
        dblMeanP = np.nan
        vecMu1 = []
        vecMu2 = []

    # trial dur
    if dblUseMaxDur is None:
        dblUseMaxDur = min(np.min(np.diff(arrEventTimes1[:, 0])), np.min(np.diff(arrEventTimes2[:, 0])))
    else:
        dblUseMaxDur = np.float64(dblUseMaxDur)
        assert dblUseMaxDur.size == 1 and dblUseMaxDur > 0, "dblUseMaxDur is not a positive scalar float"

    # get resampling num
    if intResampNum is None:
        intResampNum = np.int64(250)
    else:
        intResampNum = np.int64(intResampNum)
        assert intResampNum.size == 1 and intResampNum > 1, "intResampNum is not a positive integer"

    # plotting
    if boolPlot is None:
        boolPlot = False
    else:
        assert isinstance(boolPlot, bool), "boolPlot is not a boolean"

    # direct quantile computation
    if boolDirectQuantile is None:
        boolDirectQuantile = False
    else:
        assert isinstance(boolDirectQuantile, bool), "boolDirectQuantile is not a boolean"

    # %% check data length
    assert (len(vecTime1) == len(vecValue1) and len(vecTime2) == len(vecValue2)), 'Input lengths do not match'

    assert np.min(arrEventTimes1[:, 0]) > np.min(vecTime1) and np.max(arrEventTimes1[:, 0]) < (np.max(vecTime1)-dblUseMaxDur) \
        and np.min(arrEventTimes2[:, 0]) > np.min(vecTime2) and np.max(arrEventTimes2[:, 0]) < (np.max(vecTime2)-dblUseMaxDur), 'Events exist outside of data period'

    # %% calculate zeta
    dZETA_Two = calcTsZetaTwo(vecTime1, vecValue1, arrEventTimes1, vecTime2, vecValue2, arrEventTimes2,
                              dblSuperResFactor, dblUseMaxDur, intResampNum, boolDirectQuantile)

    # update and unpack
    dZETA.update(dZETA_Two)
    vecRefTime = dZETA['vecRefTime']
    vecRealDiff = dZETA['vecRealDiff']
    vecRealFrac1 = dZETA['vecRealFrac1']
    vecRealFrac2 = dZETA['vecRealFrac2']
    matRandDiff = dZETA['matRandDiff']
    dblZetaP = dZETA['dblZetaP']
    dblZETA = dZETA['dblZETA']
    intZETAIdx = dZETA['intZETAIdx']
    matTracePerTrial1 = dZETA['matTracePerTrial1']
    matTracePerTrial2 = dZETA['matTracePerTrial2']

    # check if calculation is valid, otherwise return empty values
    if intZETAIdx is None:
        logging.warning("zetatstest2: calculation failed, defaulting to p=1.0")
        return dblZetaP, dZETA

    # %% extract real outputs
    # get location
    dblZETATime = vecRefTime[intZETAIdx]
    dblZETADeviation = vecRealDiff[intZETAIdx]

    # find peak of inverse sign
    intZETAIdx_InvSign = np.argmax(-np.sign(dblZETADeviation)*vecRealDiff)
    dblZETATime_InvSign = vecRefTime[intZETAIdx_InvSign]
    dblZETADeviation_InvSign = vecRealDiff[intZETAIdx_InvSign]

    # %% calculate mean-rate difference
    if boolStopSupplied:
        for intTrace in [1, 2]:
            if intTrace == 1:
                # assign data
                vecEventStarts = arrEventTimes1[:, 0]
                vecEventStops = arrEventTimes1[:, 1]
                vecThisTraceT = vecTime1
                vecThisTraceAct = vecValue1
            else:
                # assign data
                vecEventStarts = arrEventTimes2[:, 0]
                vecEventStops = arrEventTimes2[:, 1]
                vecThisTraceT = vecTime2
                vecThisTraceAct = vecValue2

            intTimeNum = len(vecThisTraceT)
            intMaxRep = len(vecEventStarts)
            vecMu_Base = np.empty(intMaxRep)
            vecMu_Base.fill(np.nan)
            vecMu_Dur = np.empty(intMaxRep)
            vecMu_Dur.fill(np.nan)

            # go through trials to build mean-rate diff
            for intEvent, dblStimStartT in enumerate(vecEventStarts):
                # %% get original times
                dblStimStopT = vecEventStops[intEvent]
                dblBaseStopT = dblStimStartT + dblUseMaxDur
                if (dblBaseStopT - dblStimStopT) <= 0:
                    raise Exception(
                        "Input error: event stop times do not precede the next stimulus' start time")

                intStartT = np.max([0, findfirst(vecThisTraceT > dblStimStartT) - 1])
                intStopT = np.min([intTimeNum, findfirst(vecThisTraceT > dblStimStopT)+1])
                intEndT = np.min([intTimeNum, findfirst(vecThisTraceT > dblBaseStopT)+1])
                vecSelectFramesBase = np.arange(intStopT, intEndT)
                vecSelectFramesStim = np.arange(intStartT, intStopT)

                #  %% get data
                vecUseBaseTrace = vecThisTraceAct[vecSelectFramesBase]
                vecUseStimTrace = vecThisTraceAct[vecSelectFramesStim]

                # %% get activity
                vecMu_Base[intEvent] = np.mean(vecUseBaseTrace)
                vecMu_Dur[intEvent] = np.mean(vecUseStimTrace)

            if intTrace == 1:
                vecMu_Base1 = vecMu_Base
                vecMu_Dur1 = vecMu_Dur
            else:
                vecMu_Base2 = vecMu_Base
                vecMu_Dur2 = vecMu_Dur

        # difference
        vecMu1 = vecMu_Dur1 - vecMu_Base1
        vecMu2 = vecMu_Dur2 - vecMu_Base2

        # get metrics
        dblMeanP = stats.ttest_ind(vecMu1, vecMu2)[1]
        dblMeanZ = -stats.norm.ppf(dblMeanP/2)

    # %% build output structure
    # fill dZETA
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['dblZETADeviation'] = dblZETADeviation
    dZETA['dblZETATime'] = dblZETATime
    dZETA['intZETAIdx'] = intZETAIdx
    if boolStopSupplied:
        dZETA['dblMeanZ'] = dblMeanZ
        dZETA['dblMeanP'] = dblMeanP
        dZETA['vecMu1'] = vecMu1
        dZETA['vecMu2'] = vecMu2

    dZETA['dblZETADeviation_InvSign'] = dblZETADeviation_InvSign
    dZETA['dblZETATime_InvSign'] = dblZETATime_InvSign
    dZETA['intZETAIdx_InvSign'] = intZETAIdx_InvSign
    dZETA['vecRefTime'] = vecRefTime
    dZETA['vecRealDiff'] = vecRealDiff
    dZETA['matRandDiff'] = matRandDiff
    dZETA['vecRealFrac1'] = vecRealFrac1
    dZETA['vecRealFrac2'] = vecRealFrac2
    dZETA['dblUseMaxDur'] = dblUseMaxDur
    dZETA['matTracePerTrial1'] = matTracePerTrial1
    dZETA['matTracePerTrial2'] = matTracePerTrial2

    # %% plot
    if boolPlot:
        plottszeta2(dZETA)

    # %% return
    return dblZetaP, dZETA

# %% two-sample zeta test


def zetatest2(vecSpikeTimes1, arrEventTimes1, vecSpikeTimes2, arrEventTimes2,
              dblUseMaxDur=None, intResampNum=250, boolPlot=False, boolDirectQuantile=False):
    """
    Calculates two-sample zeta-test

    Heimel, J.A., Meijer, G.T., Montijn, J.S. (2023). A new family of statistical tests for responses
    in point-event and time-series data for one- and two-sample comparisons. bioRxiv

    Montijn, J.S., Seignette, K., Howlett, M.H., Cazemier, J.L., Kamermans, M., Levelt, C.N.,
    and Heimel, J.A. (2021). A parameter-free statistical test for neuronal responsiveness.
    eLife 10, e71969.

    Syntax:
    dblZeta2P,dZETA2 = zetatest2(vecSpikeTimes1,arrEventTimes1,vecSpikeTimes2,arrEventTimes2,
                                                  dblUseMaxDur=None, intResampNum=250, boolPlot=False, boolDirectQuantile=False)


    Parameters
    ----------
    vecSpikeTimes1 : 1D array (float)
        spike times (in seconds) for condition 1.
    arrEventTimes1 : 1D or 2D array (float)
        event on times (s) for condition 1, or [T x 2] including event off times to calculate mean-rate difference.
    vecSpikeTimes2 : 1D array (float)
        spike times (in seconds) for condition 2.
    arrEventTimes2 : 1D or 2D array (float)
        event on times (s) for condition 2, or [T x 2] including event off times to calculate mean-rate difference.

    dblUseMaxDur : float
        window length for calculating ZETA: ignore all spikes beyond this duration after event onset
        (default: minimum of event onset to event onset)
    intResampNum : integer
        number of resamplings (default: 250)
        [Note: if your p-value is close to significance, you should increase this number to enhance the precision]
    boolPlot : boolean switch
        plotting switch (False: no plot, True: plot figure) (default: False)
    boolDirectQuantile: boolean
         switch to use the empirical null-distribution rather than the Gumbel approximation (default: False)
         [Note: requires many resamplings!]

    Returns
    -------
    dblZeta2P : float
        p-value based on Zenith of Event-based Time-locked Anomalies for two-sample comparison
    dZETA2 : dict
        additional information of ZETA test
            dblZetaP; p-value based on Zenith of Event-based Time-locked Anomalies (same as above)
            dblZETA; responsiveness z-score (i.e., >2 is significant)
            dblMeanZ; z-score for mean-rate stim/base difference (i.e., >2 is significant)
            dblMeanP; p-value based on mean-rate stim/base difference
            dblZETADeviation; temporal deviation value underlying ZETA
            dblZetaT; time corresponding to ZETA
            intZetaIdx; entry corresponding to ZETA
            vecMu1; average spiking rate values per event underlying t-test for condition 1
            vecMu2; average spiking rate values per event underlying t-test for condition 2
            vecSpikeT: timestamps of spike times (corresponding to vecRealDiff)
            vecRealDiff; difference between condition 1 and 2
            vecRealFrac1; cumulative spike vector of condition 1
            vecRealFrac2; cumulative spike vector of condition 2
            dblD_InvSign; largest deviation of inverse sign to ZETA (i.e., -ZETA)
            dblZetaT_InvSign; time corresponding to -ZETA
            intZetaIdx_InvSign; entry corresponding to -ZETA
            cellRandTime; timestamps for null-hypothesis resampled data
            cellRandDiff; null-hypothesis temporal deviation vectors of resampled data
            dblUseMaxDur; window length used to calculate ZETA

    Code by Jorrit Montijn

    Version history:
    1.0 - 25 Oct 2023 Translated to python [by Jorrit Montijn]

    """

    # %% build placeholder outputs
    dblZetaP = 1.0
    dZETA = dict()

    # fill dZETA
    # ZETA significance
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = None
    # mean-rate significance
    dZETA['dblMeanZ'] = None
    dZETA['dblMeanP'] = None
    # data on ZETA peak
    dZETA['dblZETADeviation'] = None
    dZETA['dblZetaT'] = None
    dZETA['intZetaIdx'] = None
    # data underlying mean-rate test
    dZETA['vecMu1'] = None
    dZETA['vecMu2'] = None

    # inverse-sign ZETA
    dZETA['dblD_InvSign'] = None
    dZETA['dblZetaT_InvSign'] = None
    dZETA['intZetaIdx_InvSign'] = None

    # derived from calcZetaOne
    dZETA['vecSpikeT'] = None
    dZETA['vecRealDiff'] = None
    dZETA['vecRealFrac1'] = None
    dZETA['vecRealFrac2'] = None
    dZETA['cellRandTime'] = None
    dZETA['cellRandDiff'] = None
    # dZETA['dblZetaP'] = None #<-updates automatically
    # dZETA['dblZETA'] = None #<-updates automatically
    # dZETA['intZETAIdx'] = None #<-updates automatically

    # window used for analysis
    dZETA['dblUseMaxDur'] = None

    # %% prep data and assert inputs are correct

    # vecSpikeTimes1 must be [S by 1] array
    assert (len(vecSpikeTimes1.shape) == 1 or vecSpikeTimes1.shape[1] == 1) and issubclass(
        vecSpikeTimes1.dtype.type, np.floating), "Input vecSpikeTimes1 is not a 1D float np.array with >2 spike times"
    vecSpikeTimes1 = np.sort(vecSpikeTimes1.flatten(), axis=0)

    # vecSpikeTimes2 must be [S by 1] array
    assert (len(vecSpikeTimes2.shape) == 1 or vecSpikeTimes2.shape[1] == 1) and issubclass(
        vecSpikeTimes2.dtype.type, np.floating), "Input vecSpikeTimes2 is not a 1D float np.array with >2 spike times"
    vecSpikeTimes2 = np.sort(vecSpikeTimes2.flatten(), axis=0)

    # ensure orientation and assert that arrEventTimes1 is a 1D or N-by-2 array of floats
    assert len(arrEventTimes1.shape) < 3 and issubclass(
        arrEventTimes1.dtype.type, np.floating), "Input arrEventTimes1 is not a 1D or 2D float np.array"
    if len(arrEventTimes1.shape) > 1:
        if arrEventTimes1.shape[1] < 3:
            pass
        elif arrEventTimes1.shape[0] < 3:
            arrEventTimes1 = arrEventTimes1.T
        else:
            raise Exception(
                "Input error: arrEventTimes1 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes1 = np.reshape(arrEventTimes1, (-1, 1))
    # define event starts
    vecEventStarts1 = arrEventTimes1[:, 0]

    # ensure orientation and assert that arrEventTimes2 is a 1D or N-by-2 array of floats
    assert len(arrEventTimes2.shape) < 3 and issubclass(
        arrEventTimes2.dtype.type, np.floating), "Input arrEventTimes2 is not a 1D or 2D float np.array"
    if len(arrEventTimes2.shape) > 1:
        if arrEventTimes2.shape[1] < 3:
            pass
        elif arrEventTimes2.shape[0] < 3:
            arrEventTimes2 = arrEventTimes2.T
        else:
            raise Exception(
                "Input error: arrEventTimes2 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes2 = np.reshape(arrEventTimes2, (-1, 1))
    # define event starts
    vecEventStarts2 = arrEventTimes2[:, 0]

    # is stop supplied?
    if len(arrEventTimes1.shape) > 1 and arrEventTimes1.shape[1] > 1 and len(arrEventTimes2.shape) > 1 and arrEventTimes2.shape[1] > 1:
        boolStopSupplied = True
        arrEventOnDur1 = arrEventTimes1[:, 1] - arrEventTimes1[:, 0]
        assert np.all(arrEventOnDur1 > 0), "at least one event in arrEventTimes1 has a negative duration"

        arrEventOnDur2 = arrEventTimes2[:, 1] - arrEventTimes2[:, 0]
        assert np.all(arrEventOnDur2 > 0), "at least one event in arrEventTimes2 has a negative duration"

        # trial dur
        if dblUseMaxDur is None:
            dblUseMaxDur = min(np.min(arrEventOnDur1), np.min(arrEventOnDur2))
    else:
        boolStopSupplied = False
        dblMeanZ = np.nan
        dblMeanP = np.nan

    # trial dur
    if dblUseMaxDur is None:
        dblUseMaxDur = min(np.min(np.diff(arrEventTimes1[:, 0])), np.min(np.diff(arrEventTimes2[:, 0])))
    else:
        dblUseMaxDur = np.float64(dblUseMaxDur)
        assert dblUseMaxDur.size == 1 and dblUseMaxDur > 0, "dblUseMaxDur is not a positive scalar float"

    # get resampling num
    if intResampNum is None:
        intResampNum = np.int64(250)
    else:
        intResampNum = np.int64(intResampNum)
        assert intResampNum.size == 1 and intResampNum > 1, "intResampNum is not a positive integer"

    # plotting
    if boolPlot is None:
        boolPlot = False
    else:
        assert isinstance(boolPlot, bool), "boolPlot is not a boolean"

    # direct quantile comnputation
    if boolDirectQuantile is None:
        boolDirectQuantile = False
    else:
        assert isinstance(boolDirectQuantile, bool), "boolDirectQuantile is not a boolean"

    # to do: parallel computing
    boolParallel = False

    # %% calculate zeta
    vecEventStarts1 = arrEventTimes1[:, 0]
    vecEventStarts2 = arrEventTimes2[:, 0]
    # if len(vecEventStarts1) > 1 and (len(vecSpikeTimes1)+len(vecSpikeTimes2)) > 0 and dblUseMaxDur is not None and dblUseMaxDur > 0:
    dZETA_Two = calcZetaTwo(vecSpikeTimes1, vecEventStarts1, vecSpikeTimes2,
                            vecEventStarts2, dblUseMaxDur, intResampNum, boolDirectQuantile)

    # %% calculate zeta
    # update and unpack
    dZETA.update(dZETA_Two)
    vecSpikeT = dZETA['vecSpikeT']
    vecRealDiff = dZETA['vecRealDiff']
    dblZetaP = dZETA['dblZetaP']
    intZetaIdx = dZETA['intZETAIdx']

    # check if calculation is valid, otherwise return empty values
    if intZetaIdx is None:
        logging.warning("zetatest2: calculation failed, defaulting to p=1.0")
        return dblZetaP, dZETA

    # %% extract real outputs
    # get location
    dblZetaT = vecSpikeT[intZetaIdx]
    dblZETADeviation = vecRealDiff[intZetaIdx]

    # find peak of inverse sign
    intZetaIdx_InvSign = np.argmax(-np.sign(dblZETADeviation)*vecRealDiff)
    dblZetaT_InvSign = vecSpikeT[intZetaIdx_InvSign]
    dblD_InvSign = vecRealDiff[intZetaIdx_InvSign]

    # %% calculate mean-rate difference with t-test
    if boolStopSupplied:
        # calculate spike counts and durations during baseline and stimulus times
        vecRespBinsDur = np.sort(np.reshape(arrEventTimes1, -1))
        vecR, arrBins = np.histogram(vecSpikeTimes1, bins=vecRespBinsDur)
        vecD = np.diff(vecRespBinsDur)
        vecMu1 = np.divide(np.float64(vecR[0:len(vecR):2]), vecD[0:len(vecD):2])

        # calculate mean rates during off-times
        vecRespBinsDur = np.sort(np.reshape(arrEventTimes2, -1))
        vecR, arrBins = np.histogram(vecSpikeTimes2, bins=vecRespBinsDur)
        vecD = np.diff(vecRespBinsDur)
        vecMu2 = np.divide(np.float64(vecR[0:len(vecR):2]), vecD[0:len(vecD):2])

        # get metrics
        dblMeanP = stats.ttest_ind(vecMu1, vecMu2)[1]
        dblMeanZ = -stats.norm.ppf(dblMeanP/2)

    # %% build output dictionary
    # fill dZETA
    dZETA['dblZETADeviation'] = dblZETADeviation
    dZETA['dblZetaT'] = dblZetaT
    if boolStopSupplied:
        dZETA['dblMeanZ'] = dblMeanZ
        dZETA['dblMeanP'] = dblMeanP
        dZETA['vecMu1'] = vecMu1
        dZETA['vecMu2'] = vecMu2

    # inverse-sign ZETA
    dZETA['dblD_InvSign'] = dblD_InvSign
    dZETA['dblZetaT_InvSign'] = dblZetaT_InvSign
    dZETA['intZetaIdx_InvSign'] = intZetaIdx_InvSign
    # window used for analysis
    dZETA['dblUseMaxDur'] = dblUseMaxDur

    # %% plot
    if boolPlot:
        plotzeta2(vecSpikeTimes1, vecEventStarts1, vecSpikeTimes2, vecEventStarts2, dZETA)

    # %% return outputs
    return dblZetaP, dZETA

# %% time-series zeta


def zetatstest(vecTime, vecValue, arrEventTimes,
               dblUseMaxDur=None, intResampNum=100, boolPlot=False, dblJitterSize=2.0, boolDirectQuantile=False, boolStitch=True):
    """
    Calculates responsiveness index zeta for timeseries data

    Montijn, J.S., Seignette, K., Howlett, M.H., Cazemier, J.L., Kamermans, M., Levelt, C.N.,
    and Heimel, J.A. (2021). A parameter-free statistical test for neuronal responsiveness.
    eLife 10, e71969.

    Montijn J.S. and Heimel J.A. (2022). A novel and highly sensitive statistical test for calcium imaging.
   FENS meeting 2022, Poster S03-480

    Syntax:
    dblZetaP,dZETA = zetatstest(vecTime, vecValue, arrEventTimes,
                 dblUseMaxDur=None, intResampNum=100, boolPlot=False, dblJitterSize=2.0, boolDirectQuantile=False, boolStitch=True):

    Parameters
    ----------
    vecTime [N x 1]: 1D array (float)
        timestamps in seconds corresponding to entries in vecValue.
    vecValue [N x 1] : 1D array (float)
        values (e.g., dF/F0 activity).
    arrEventTimes : 1D or 2D array (float)
        event on times (s), or [T x 2] including event off times to calculate mean-rate difference.
    dblUseMaxDur : float
        window length for calculating ZETA: ignore all entries beyond this duration after event onset
        (default: minimum of event onset to event onset)
    intResampNum : integer
        number of resamplings (default: 100)
        [Note: if your p-value is close to significance, you should increase this number to enhance the precision]
    boolPlot : boolean switch
        plotting switch (False: no plot, True: plot figure) (default: False)
    dblJitterSize; float
        sets the temporal jitter window relative to dblUseMaxDur (default: 2.0)
    boolDirectQuantile: boolean
         switch to use the empirical null-distribution rather than the Gumbel approximation (default: False)
         [Note: requires many resamplings!]
    boolStitch: boolean
        switch to perform data stitching (default: True)


    Returns
    -------
    dblZetaP : float
        p-value based on Zenith of Event-based Time-locked Anomalies
    dZETA : dict
        additional information of ZETA test
            dblZetaP; p-value based on Zenith of Event-based Time-locked Anomalies (same as above)
            dblZETA; responsiveness z-score (i.e., >2 is significant)
            dblMeanZ; z-score for mean-rate stim/base difference (i.e., >2 is significant)
            dblMeanP; p-value based on mean-rate stim/base difference
            dblZETADeviation; temporal deviation value underlying ZETA
            dblLatencyZETA; time corresponding to ZETA
            intZETAIdx; entry corresponding to ZETA
            vecMu_Dur; mean activity per trial during stim (used for mean-rate test)
            vecMu_Base; mean activity per trial during baseline (used for mean-rate test)
            dblD_InvSign; largest deviation of inverse sign to ZETA (i.e., -ZETA)
            dblLatencyInvZETA; time corresponding to -ZETA
            intIdx_InvSign; entry corresponding to -ZETA
            vecRealTime: timestamps of event-centered time-series values (corresponding to vecRealDeviation)
            vecRealDeviation; temporal deviation vector of data
            vecRealFrac; cumulative distribution of spike times
            vecRealFracLinear; linear baseline of cumulative distribution
            matRandDeviation; baseline temporal deviation matrix of jittered data
            dblUseMaxDur; window length used to calculate ZETA


    Code by Jorrit Montijn

    Version history:
    1.0 - 24 August 2023 Translation to python by Jorrit Montijn
    1.1 - 25 Oct 2023 Removed jitter distro switch [by Jorrit Montijn]

    """
    # %% build placeholder outputs
    dblZetaP = 1.0
    dZETA = dict()

    # fill dZETA
    # ZETA significance
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = None
    # mean-rate significance
    dZETA['dblMeanZ'] = None
    dZETA['dblMeanP'] = None
    # data on ZETA peak
    dZETA['dblZETADeviation'] = None
    dZETA['dblLatencyZETA'] = None
    dZETA['intZETAIdx'] = None
    # data underlying mean-rate test
    dZETA['vecMu_Dur'] = None
    dZETA['vecMu_Base'] = None
    # inverse-sign ZETA
    dZETA['dblD_InvSign'] = None
    dZETA['dblLatencyInvZETA'] = None
    dZETA['intIdx_InvSign'] = None

    # derived from calcZetaOne
    dZETA['vecRealTime'] = None
    dZETA['vecRealDeviation'] = None
    dZETA['vecRealFrac'] = None
    dZETA['vecRealFracLinear'] = None
    dZETA['cellRandTime'] = None
    dZETA['cellRandDeviation'] = None
    # dZETA['dblZetaP'] = None #<-updates automatically
    # dZETA['dblZETA'] = None #<-updates automatically
    # dZETA['intZETAIdx'] = None #<-updates automatically

    # window used for analysis
    dZETA['dblUseMaxDur'] = None

    # %% prep data and assert inputs are correct

    # vecTime and vecValue must be [N by 1] arrays
    assert len(vecTime.shape) == len(
        vecValue.shape) and vecTime.shape == vecValue.shape, "vecTime and vecValue have different shapes"
    assert (len(vecTime.shape) == 1 or vecTime.shape[1] == 1) and issubclass(
        vecTime.dtype.type, np.floating), "Input vecTime is not a 1D float np.array with >2 spike times"
    vecTime = vecTime.flatten()
    vecValue = vecValue.flatten()
    vecReorder = np.argsort(vecTime, axis=0)
    vecTime = vecTime[vecReorder]
    vecValue = vecValue[vecReorder]

    # ensure orientation and assert that arrEventTimes is a 1D or N-by-2 array of floats
    assert len(arrEventTimes.shape) < 3 and issubclass(
        arrEventTimes.dtype.type, np.floating), "Input arrEventTimes is not a 1D or 2D float np.array"
    if len(arrEventTimes.shape) > 1:
        if arrEventTimes.shape[1] < 3:
            pass
        elif arrEventTimes.shape[0] < 3:
            arrEventTimes = arrEventTimes.T
        else:
            raise Exception(
                "Input error: arrEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes = np.reshape(arrEventTimes, (-1, 1))
    # define event starts
    vecEventStarts = arrEventTimes[:, 0]

    # check if number of events and values is sufficient
    if vecTime.size < 3 or vecEventStarts.size < 3:
        if vecTime.size < 3:
            strMsg1 = f"Number of entries in time-series ({vecTime.size}) is too few to calculate zeta; "
        else:
            strMsg1 = ""
        if vecEventStarts.size < 3:
            strMsg2 = f"Number of events ({vecEventStarts.size}) is too few to calculate zeta; "
        else:
            strMsg2 = ""
        logging.warning("zetatstest: " + strMsg1 + strMsg2 + "defaulting to p=1.0")

        return dblZetaP, dZETA

    # is stop supplied?
    if len(arrEventTimes.shape) > 1 and arrEventTimes.shape[1] > 1:
        boolStopSupplied = True
        vecEventStops = arrEventTimes[:, 1]
        vecEventOnDur = arrEventTimes[:, 1] - arrEventTimes[:, 0]
        assert np.all(vecEventOnDur > 0), "at least one event in arrEventTimes has a non-positive duration"
        vecMu_Dur = np.zeros(vecEventStops.shape)
        vecMu_Base = np.zeros(vecEventStops.shape)

    else:
        boolStopSupplied = False
        dblMeanZ = np.nan
        dblMeanP = np.nan
        vecMu_Dur = []
        vecMu_Base = []

    # trial dur
    if dblUseMaxDur is None:
        dblUseMaxDur = np.min(np.diff(arrEventTimes[:, 0]))
    else:
        dblUseMaxDur = np.float64(dblUseMaxDur)
        assert dblUseMaxDur.size == 1 and dblUseMaxDur > 0, "dblUseMaxDur is not a positive scalar float"

    # get resampling num
    if intResampNum is None:
        intResampNum = np.int64(100)
    else:
        intResampNum = np.int64(intResampNum)
        assert intResampNum.size == 1 and intResampNum > 1, "intResampNum is not a positive integer"

    # plotting
    if boolPlot is None:
        boolPlot = False
    else:
        assert isinstance(boolPlot, bool), "boolPlot is not a boolean"

    # jitter
    if dblJitterSize is None:
        dblJitterSize = np.float64(2.0)
    else:
        dblJitterSize = np.float64(dblJitterSize)
        assert dblJitterSize.size == 1 and dblJitterSize > 0, "dblJitterSize is not a postive scalar float"

    # direct quantile computation
    if boolDirectQuantile is None:
        boolDirectQuantile = False
    else:
        assert isinstance(boolDirectQuantile, bool), "boolDirectQuantile is not a boolean"

    # stitch?
    if boolStitch is None:
        boolStitch = True
    else:
        assert isinstance(boolStitch, bool), "boolStitch is not a boolean"

    # %% check data length
    dblDataT0 = np.min(vecTime)
    dblReqT0 = np.min(vecEventStarts) - dblJitterSize*dblUseMaxDur
    if dblDataT0 > dblReqT0:
        logging.warning("zetatstest: leading data preceding first event is insufficient for maximal jittering")

    dblDataT_end = np.max(vecTime)
    dblReqT_end = np.max(vecEventStarts) + dblJitterSize*dblUseMaxDur + dblUseMaxDur
    if dblDataT_end < dblReqT_end:
        logging.warning("zetatstest: lagging data after last event is insufficient for maximal jittering")

    # %% calculate zeta
    dZETA_One = calcTsZetaOne(vecTime, vecValue, vecEventStarts, dblUseMaxDur, intResampNum,
                              boolDirectQuantile, dblJitterSize, boolStitch)

    # update and unpack
    dZETA.update(dZETA_One)
    vecRealTime = dZETA['vecRealTime']
    vecRealDeviation = dZETA['vecRealDeviation']
    vecRealFrac = dZETA['vecRealFrac']
    vecRealFracLinear = dZETA['vecRealFracLinear']
    cellRandTime = dZETA['cellRandTime']
    cellRandDeviation = dZETA['cellRandDeviation']
    dblZetaP = dZETA['dblZetaP']
    dblZETA = dZETA['dblZETA']
    intZETAIdx = dZETA['intZETAIdx']

    # check if calculation is valid, otherwise return empty values
    if intZETAIdx is None:
        logging.warning("zetatstest: calculation failed, defaulting to p=1.0")
        return dblZetaP, dZETA

    # %% extract real outputs
    # get location
    dblLatencyZETA = vecRealTime[intZETAIdx]
    dblZETADeviation = vecRealDeviation[intZETAIdx]

    # find peak of inverse sign
    intIdx_InvSign = np.argmax(-np.sign(dblZETADeviation)*vecRealDeviation)
    dblLatencyInvZETA = vecRealTime[intIdx_InvSign]
    dblD_InvSign = vecRealDeviation[intIdx_InvSign]

    # %% calculate mean-rate difference
    if boolStopSupplied:
        # pre-allocate
        intTimeNum = len(vecTime)-1

        # go through trials to build spike time vector
        for intEvent, dblStimStartT in enumerate(vecEventStarts):
            # %% get original times
            dblStimStopT = vecEventStops[intEvent]
            dblBaseStopT = dblStimStartT + dblUseMaxDur
            if (dblBaseStopT - dblStimStopT) <= 0:
                raise Exception(
                    "Input error: event stop times do not precede the next stimulus' start time")

            intStartT = np.max([0, findfirst(vecTime > dblStimStartT) - 1])
            intStopT = np.min([intTimeNum, findfirst(vecTime > dblStimStopT)+1])
            intEndT = np.min([intTimeNum, findfirst(vecTime > dblBaseStopT)+1])
            vecSelectFramesBase = np.arange(intStopT, intEndT)
            vecSelectFramesStim = np.arange(intStartT, intStopT)

           #  %% get data
            vecUseBaseTrace = vecValue[vecSelectFramesBase]
            vecUseStimTrace = vecValue[vecSelectFramesStim]

            # %% get activity
            vecMu_Base[intEvent] = np.mean(vecUseBaseTrace)
            vecMu_Dur[intEvent] = np.mean(vecUseStimTrace)

        # get metrics
        indUseTrials = np.logical_and(~np.isnan(vecMu_Dur), ~np.isnan(vecMu_Base))
        vecMu_Dur = vecMu_Dur[indUseTrials]
        vecMu_Base = vecMu_Base[indUseTrials]
        dblMeanP = stats.ttest_rel(vecMu_Dur, vecMu_Base)[1]
        dblMeanZ = -stats.norm.ppf(dblMeanP/2)

    # %% build output structure
    # fill dZETA
    dZETA['dblZETADeviation'] = dblZETADeviation
    dZETA['dblLatencyZETA'] = dblLatencyZETA
    if boolStopSupplied:
        dZETA['dblMeanZ'] = dblMeanZ
        dZETA['dblMeanP'] = dblMeanP
        dZETA['vecMu_Dur'] = vecMu_Dur
        dZETA['vecMu_Base'] = vecMu_Base

    # inverse-sign ZETA
    dZETA['dblD_InvSign'] = dblD_InvSign
    dZETA['dblLatencyInvZETA'] = dblLatencyInvZETA
    dZETA['intIdx_InvSign'] = intIdx_InvSign
    # window used for analysis
    dZETA['dblUseMaxDur'] = dblUseMaxDur

    # %% plot
    if boolPlot:
        plottszeta(vecTime, vecValue, vecEventStarts, dZETA)

    # %% return
    return dblZetaP, dZETA

# %% zetatest


def zetatest(vecSpikeTimes, arrEventTimes,
             dblUseMaxDur=None, intResampNum=100, boolPlot=False, dblJitterSize=2.0,
             tplRestrictRange=(-np.inf, np.inf), boolStitch=True,
             boolDirectQuantile=False, boolReturnRate=False):
    """
    Calculates neuronal responsiveness index ZETA.

    Montijn, J.S., Seignette, K., Howlett, M.H., Cazemier, J.L., Kamermans, M., Levelt, C.N.,
    and Heimel, J.A. (2021). A parameter-free statistical test for neuronal responsiveness.
    eLife 10, e71969.

    Syntax:
    dblZetaP,dZETA,dRate,vecLatencies = zetatest(vecSpikeTimes,arrEventTimes,
                                                   dblUseMaxDur=None, intResampNum=100, boolPlot=False, dblJitterSize=2.0,
                                                   tplRestrictRange=(-np.inf, np.inf), boolStitch=True,
                                                   boolDirectQuantile=False, boolReturnRate=False):

    Parameters
    ----------
    vecSpikeTimes : 1D array (float)
        spike times (in seconds).
    arrEventTimes : 1D or 2D array (float)
        event on times (s), or [T x 2] including event off times to calculate mean-rate difference.

    dblUseMaxDur : float
        window length for calculating ZETA: ignore all spikes beyond this duration after event onset
        (default: minimum of event onset to event onset)
    intResampNum : integer
        number of resamplings (default: 100)
        [Note: if your p-value is close to significance, you should increase this number to enhance the precision]
    boolPlot : boolean switch
        plotting switch (False: no plot, True: plot figure) (default: False)
    dblJitterSize; float
        sets the temporal jitter window relative to dblUseMaxDur (default: 2.0)
    tplRestrictRange : 2-element tuple
        temporal range within which to restrict onset/peak latencies (default: [-inf inf])
    boolStitch; boolean
        switch to use data-stitching to ensure continuous time (default: True)
    boolDirectQuantile: boolean
         switch to use the empirical null-distribution rather than the Gumbel approximation (default: False)
         [Note: requires many resamplings!]
    boolReturnRate : boolean
        switch to return dictionary with spiking rate features [note: return-time is much faster if this is False]

    Returns
    -------
    dblZetaP : float
        p-value based on Zenith of Event-based Time-locked Anomalies
    dZETA : dict
        additional information of ZETA test
            dblZetaP; p-value based on Zenith of Event-based Time-locked Anomalies (same as above)
            dblZETA; responsiveness z-score (i.e., >2 is significant)
            dblMeanZ; z-score for mean-rate stim/base difference (i.e., >2 is significant)
            dblMeanP; p-value based on mean-rate stim/base difference
            dblZETADeviation; temporal deviation value underlying ZETA
            dblLatencyZETA; time corresponding to ZETA
            intZETAIdx; entry corresponding to ZETA
            vecMu_Dur; spiking rate per trial during stim (used for mean-rate test)
            vecMu_Pre; spiking rate per trial during baseline (used for mean-rate test)
            dblD_InvSign; largest deviation of inverse sign to ZETA (i.e., -ZETA)
            dblLatencyInvZETA; time corresponding to -ZETA
            intIdx_InvSign; entry corresponding to -ZETA
            vecSpikeT: timestamps of spike times (corresponding to vecRealDeviation)
            vecRealDeviation; temporal deviation vector of data
            vecRealFrac; cumulative distribution of spike times
            vecRealFracLinear; linear baseline of cumulative distribution
            cellRandTime; jittered spike times corresponding to cellRandDeviation
            cellRandDeviation; baseline temporal deviation matrix of jittered data
            dblUseMaxDur; window length used to calculate ZETA
            vecLatencies; 4-element array with latency times for different events:
                1) Latency of ZETA [same as dblZETADeviation]
                2) Latency of largest z-score with inverse sign to ZETA (same as dblLatencyInvZETA])
                3) Peak time of instantaneous firing rate (same as dRate['dblLatencyPeak'])
                4) Onset time, defined as the first crossing of peak half-height (same as dRate['dblLatencyPeakOnset'])
            vecLatencyVals; values corresponding to above latencies (ZETA, -ZETA, rate at peak, rate at onset)

    dRate : dict (empty if boolReturnRate was not set to True)
        additional parameters of the firing rate, return with boolReturnRate
            vecRate; instantaneous spiking rates (like a PSTH)
            vecT; time-points corresponding to vecRate (same as dZETA.vecSpikeT)
            vecM; Mean of multi-scale derivatives
            vecScale; timescales used to calculate derivatives
            matMSD; multi-scale derivatives matrix
            vecV; values on which vecRate is calculated (same as dZETA.vecZ)
        Data on the peak and onset:
            dblLatencyPeak; time of peak (in seconds) [vecLatencies entry #3]
            dblPeakWidth; duration of peak (in seconds) [vecLatencies entry #3]
            vecPeakStartStop; start and stop time of peak (in seconds) [vecLatencies entry #3]
            intPeakLoc; spike index of peak (corresponding to dZETA.vecSpikeT) [vecLatencies entry #3]
            vecPeakStartStopIdx; spike indices of peak start/stop (corresponding to dZETA.vecSpikeT) [vecLatencies entry #3]
            dblLatencyPeakOnset: latency for peak onset [vecLatencies entry #4]
    Code by Jorrit Montijn, Guido Meijer & Alexander Heimel

    Version history:
    2.5 - 17 June 2020 Translated to python [Alexander Heimel]
    2.5.1 - 18 February 2022 Bugfix of 1D arrEventTimes [by Guido Meijer]
    2.6 - 20 February 2022 Refactoring of python code [by Guido Meijer]
    3.0 - 16 Aug 2023 New port of zetatest to Python [by Jorrit Montijn]
    3.1 - 15 Sept 2023 Changed latency variable names, added intUseJitterDistro switch [by Jorrit Montijn]
    3.2 - 25 Oct 2023 Removed intUseJitterDistro switch [by Jorrit Montijn]

    """

    # %% build placeholder outputs
    dblZetaP = 1.0
    dZETA = dict()
    dRate = dict()
    vecLatencies = np.empty((4, 1))
    vecLatencies.fill(np.nan)
    vecLatencyVals = vecLatencies

    # fill dZETA
    # ZETA significance
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = None
    # mean-rate significance
    dZETA['dblMeanZ'] = None
    dZETA['dblMeanP'] = None
    # data on ZETA peak
    dZETA['dblZETADeviation'] = None
    dZETA['dblLatencyZETA'] = None
    dZETA['intZETAIdx'] = None
    # data underlying mean-rate test
    dZETA['vecMu_Dur'] = None
    dZETA['vecMu_Pre'] = None
    # inverse-sign ZETA
    dZETA['dblD_InvSign'] = None
    dZETA['dblLatencyInvZETA'] = None
    dZETA['intIdx_InvSign'] = None

    # derived from calcZetaOne
    dZETA['vecSpikeT'] = None
    dZETA['vecRealDeviation'] = None
    dZETA['vecRealFrac'] = None
    dZETA['vecRealFracLinear'] = None
    dZETA['cellRandTime'] = None
    dZETA['cellRandDeviation'] = None
    # dZETA['dblZetaP'] = None #<-updates automatically
    # dZETA['dblZETA'] = None #<-updates automatically
    # dZETA['intZETAIdx'] = None #<-updates automatically

    # window used for analysis
    dZETA['dblUseMaxDur'] = None
    # copy of latency vectors
    dZETA['vecLatencies'] = vecLatencies
    dZETA['vecLatencyVals'] = vecLatencyVals

    # fill dRate
    dRate['vecRate'] = None
    dRate['vecT'] = None
    dRate['vecM'] = None
    dRate['vecScale'] = None
    dRate['matMSD'] = None
    dRate['vecV'] = None
    dRate['dblLatencyPeak'] = None
    dRate['dblPeakWidth'] = None
    dRate['vecPeakStartStop'] = None
    dRate['intPeakLoc'] = None
    dRate['vecPeakStartStopIdx'] = None
    dRate['dblLatencyPeakOnset'] = None

    # %% prep data and assert inputs are correct

    # vecSpikeTimes must be [S by 1] array
    assert (len(vecSpikeTimes.shape) == 1 or vecSpikeTimes.shape[1] == 1) and issubclass(
        vecSpikeTimes.dtype.type, np.floating), "Input vecSpikeTimes is not a 1D float np.array with >2 spike times"
    vecSpikeTimes = np.sort(vecSpikeTimes.flatten(), axis=0)

    # ensure orientation and assert that arrEventTimes is a 1D or N-by-2 array of floats
    assert len(arrEventTimes.shape) < 3 and issubclass(
        arrEventTimes.dtype.type, np.floating), "Input arrEventTimes is not a 1D or 2D float np.array"
    if len(arrEventTimes.shape) > 1:
        if arrEventTimes.shape[1] < 3:
            pass
        elif arrEventTimes.shape[0] < 3:
            arrEventTimes = arrEventTimes.T
        else:
            raise Exception(
                "Input error: arrEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes = np.reshape(arrEventTimes, (-1, 1))
    # define event starts
    vecEventStarts = arrEventTimes[:, 0]

    # check if number of events and spikes is sufficient
    if vecSpikeTimes.size < 3 or vecEventStarts.size < 3:
        if vecSpikeTimes.size < 3:
            strMsg1 = f"Number of spikes ({vecSpikeTimes.size}) is too few to calculate zeta; "
        else:
            strMsg1 = ""
        if vecEventStarts.size < 3:
            strMsg2 = f"Number of events ({vecEventStarts.size}) is too few to calculate zeta; "
        else:
            strMsg2 = ""
        logging.warning("zetatest: " + strMsg1 + strMsg2 + "defaulting to p=1.0")

        return dblZetaP, dZETA, dRate, vecLatencies

    # is stop supplied?
    if len(arrEventTimes.shape) > 1 and arrEventTimes.shape[1] > 1:
        boolStopSupplied = True
        arrEventOnDur = arrEventTimes[:, 1] - arrEventTimes[:, 0]
        assert np.all(arrEventOnDur > 0), "at least one event in arrEventTimes has a negative duration"

    else:
        boolStopSupplied = False
        dblMeanZ = np.nan
        dblMeanP = np.nan

    # trial dur
    if dblUseMaxDur is None:
        dblUseMaxDur = np.min(np.diff(arrEventTimes[:, 0]))
    else:
        dblUseMaxDur = np.float64(dblUseMaxDur)
        assert dblUseMaxDur.size == 1 and dblUseMaxDur > 0, "dblUseMaxDur is not a positive scalar float"

    # get resampling num
    if intResampNum is None:
        intResampNum = np.int64(100)
    else:
        intResampNum = np.int64(intResampNum)
        assert intResampNum.size == 1 and intResampNum > 1, "intResampNum is not a positive integer"

    # plotting
    if boolPlot is None:
        boolPlot = False
    else:
        assert isinstance(boolPlot, bool), "boolPlot is not a boolean"

    # jitter
    if dblJitterSize is None:
        dblJitterSize = np.float64(2.0)
    else:
        dblJitterSize = np.float64(dblJitterSize)
        assert dblJitterSize.size == 1 and dblJitterSize > 0, "dblJitterSize is not a postive scalar float"

    # restrict range
    if tplRestrictRange is None:
        tplRestrictRange = np.float64((-np.inf, np.inf))
    else:
        tplRestrictRange = np.float64(tplRestrictRange)
        assert tplRestrictRange.size == 2, "tplRestrictRange does not have two values"

    # stitching
    if boolStitch is None:
        boolStitch = True
    else:
        assert isinstance(boolStitch, bool), "boolStitch is not a boolean"

    # direct quantile comnputation
    if boolDirectQuantile is None:
        boolDirectQuantile = False
    else:
        assert isinstance(boolDirectQuantile, bool), "boolDirectQuantile is not a boolean"

    # return dRate
    if boolReturnRate is None:
        boolReturnRate = False
    else:
        assert isinstance(boolReturnRate, bool), "boolReturnRate is not a boolean"
    if boolPlot is True and boolReturnRate is False:
        boolReturnRate = True
        logging.warning(
            "zetatest: boolReturnRate was False, but you requested plotting, so boolReturnRate is now set to True")

    # to do: parallel computing
    boolParallel = False

    # %% calculate zeta
    dZETA_One = calcZetaOne(vecSpikeTimes, vecEventStarts, dblUseMaxDur, intResampNum,
                            boolDirectQuantile, dblJitterSize, boolStitch, boolParallel)

    # update and unpack
    dZETA.update(dZETA_One)
    vecSpikeT = dZETA['vecSpikeT']
    vecRealDeviation = dZETA['vecRealDeviation']
    vecRealFrac = dZETA['vecRealFrac']
    vecRealFracLinear = dZETA['vecRealFracLinear']
    cellRandTime = dZETA['cellRandTime']
    cellRandDeviation = dZETA['cellRandDeviation']
    dblZetaP = dZETA['dblZetaP']
    dblZETA = dZETA['dblZETA']
    intZETAIdx = dZETA['intZETAIdx']

    # check if calculation is valid, otherwise return empty values
    if intZETAIdx is None:
        logging.warning("zetatest: calculation failed, defaulting to p=1.0")
        return dblZetaP, dZETA, dRate

    # %% extract real outputs
    # get location
    dblLatencyZETA = vecSpikeT[intZETAIdx]
    dblZETADeviation = vecRealDeviation[intZETAIdx]

    # find peak of inverse sign
    intIdx_InvSign = np.argmax(-np.sign(dblZETADeviation)*vecRealDeviation)
    dblLatencyInvZETA = vecSpikeT[intIdx_InvSign]
    dblD_InvSign = vecRealDeviation[intIdx_InvSign]

    # %% calculate mean-rate difference with t-test
    if boolStopSupplied:
        # calculate spike counts and durations during baseline and stimulus times
        vecRespBinsDur = np.sort(np.reshape(arrEventTimes, -1))
        vecR, arrBins = np.histogram(vecSpikeTimes, bins=vecRespBinsDur)
        vecD = np.diff(vecRespBinsDur)

        # mean rate during on-time
        vecMu_Dur = np.divide(np.float64(vecR[0:len(vecR):2]), vecD[0:len(vecD):2])

        # calculate mean rates during off-times
        dblStart1 = np.min(vecRespBinsDur)
        dblFirstPreDur = dblStart1 - np.max(dblStart1 - np.median(vecD[1:len(vecD):2]), initial=0) + np.finfo(float).eps
        dblR1 = np.sum(np.logical_and(vecSpikeTimes > (dblStart1 - dblFirstPreDur), vecSpikeTimes < dblStart1))
        vecCounts = np.concatenate([[dblR1], vecR[1:len(vecR):2]])
        vecDurs = np.concatenate([[dblFirstPreDur], vecD[1:len(vecD):2]])
        vecMu_Pre = np.divide(vecCounts, vecDurs)

        # get metrics
        dblMeanP = stats.ttest_rel(vecMu_Dur, vecMu_Pre)[1]
        dblMeanZ = -stats.norm.ppf(dblMeanP/2)

    # %% calculate instantaneous firing rates
    if boolReturnRate:
        # get average of multi-scale derivatives, and rescaled to instantaneous spiking rate
        dblMeanRate = vecSpikeT.size/(dblUseMaxDur*vecEventStarts.size)
        vecRate, dRate = getMultiScaleDeriv(vecSpikeT, vecRealDeviation,
                                            dblMeanRate=dblMeanRate, dblUseMaxDur=dblUseMaxDur, boolParallel=boolParallel)

        # %% calculate IFR statistics
        if vecRate is not None:
            # get IFR peak
            dPeak = getPeak(vecRate, dRate['vecT'], tplRestrictRange=tplRestrictRange)
            dRate.update(dPeak)
            if dRate['dblLatencyPeak'] is not None and ~np.isnan(dRate['dblLatencyPeak']):
                # assign array data
                intZetaIdxRate = min(max(0, intZETAIdx-1), len(vecRate)-1)
                intZetaIdxInvRate = min(max(0, intIdx_InvSign-1), len(vecRate)-1)

                # get onset
                dOnset = getOnset(vecRate, dRate['vecT'], dRate['dblLatencyPeak'], tplRestrictRange)
                dRate['dblLatencyPeakOnset'] = dOnset['dblLatencyPeakOnset']
                vecLatencies = [dblLatencyZETA, dblLatencyInvZETA,
                                dRate['dblLatencyPeak'], dOnset['dblLatencyPeakOnset']]
                vecLatencyVals = [vecRate[intZetaIdxRate], vecRate[intZetaIdxInvRate],
                                  vecRate[dPeak['intPeakLoc']], dOnset['dblValue']]

    # %% build output dictionary
    # fill dZETA
    dZETA['dblZETADeviation'] = dblZETADeviation
    dZETA['dblLatencyZETA'] = dblLatencyZETA
    if boolStopSupplied:
        dZETA['dblMeanZ'] = dblMeanZ
        dZETA['dblMeanP'] = dblMeanP
        dZETA['vecMu_Dur'] = vecMu_Dur
        dZETA['vecMu_Pre'] = vecMu_Pre

    # inverse-sign ZETA
    dZETA['dblD_InvSign'] = dblD_InvSign
    dZETA['dblLatencyInvZETA'] = dblLatencyInvZETA
    dZETA['intIdx_InvSign'] = intIdx_InvSign
    # window used for analysis
    dZETA['dblUseMaxDur'] = dblUseMaxDur
    # copy of latency vectors
    dZETA['vecLatencies'] = vecLatencies
    dZETA['vecLatencyVals'] = vecLatencyVals

    # %% plot
    if boolPlot:
        plotzeta(vecSpikeTimes, vecEventStarts, dZETA, dRate)

    # %% return outputs
    return dblZetaP, dZETA, dRate

# %% IFR


def ifr(vecSpikeTimes, vecEventTimes,
        dblUseMaxDur=None, dblSmoothSd=2.0, dblMinScale=None, dblBase=1.5, boolParallel=False):
    """Returns instantaneous firing rates. Syntax:
        ifr(vecSpikeTimes,vecEventTimes,
               dblUseMaxDur=None, dblSmoothSd=2, dblMinScale=None, dblBase=1.5)

    Required input:
        - vecSpikeTimes [S x 1]: spike times (s)
        - vecEventTimes [T x 1]: event on times (s), or [T x 2] including event off times

    Optional inputs:
        - dblUseMaxDur: float (s), ignore all spikes beyond this duration after stimulus onset
                                    [default: median of trial start to trial start]
        - dblSmoothSd: float, Gaussian SD of smoothing kernel (in # of bins) [default: 2]
        - dblMinScale: minimum derivative scale in seconds [default: round(log(1/1000) / log(dblBase))]
        - dblBase: critical value for locally dynamic derivative [default: 1.5]

    Outputs:
        - vecTime: array with timestamps
        - vecRate: array with instantaneous firing rates
        - dIFR; dictionary with entries:
            - vecTime: same as first output
            - vecRate: same as second output
            - vecDeviation: ZETA's deviation vector
            - vecScale: temporal scales used to calculate the multi-scale derivative

    Version history:
    1.0 - June 24, 2020 Created by Jorrit Montijn, translated to python by Alexander Heimel
    2.0 - August 22, 2023 Updated translation [by JM]
   """
    # %% prep data and assert inputs are correct
    # pre-allocate outputs
    vecTime = np.empty(0)
    vecRate = np.empty(0)
    dIFR = dict()
    dIFR['vecTime'] = vecTime
    dIFR['vecRate'] = vecRate
    dIFR['vecDeviation'] = np.empty(0)
    dIFR['vecScale'] = np.empty(0)

    # vecSpikeTimes must be [S by 1] array
    assert (len(vecSpikeTimes.shape) == 1 or vecSpikeTimes.shape[1] == 1) and issubclass(
        vecSpikeTimes.dtype.type, np.floating), "Input vecSpikeTimes is not a 1D float np.array with >2 spike times"
    vecSpikeTimes = np.sort(vecSpikeTimes.flatten(), axis=0)

    # ensure orientation and assert that arrEventTimes is a 1D or N-by-2 array of floats
    assert len(vecEventTimes.shape) < 3 and issubclass(
        vecEventTimes.dtype.type, np.floating), "Input vecEventTimes is not a 1D or 2D float np.array"
    if len(vecEventTimes.shape) > 1:
        if vecEventTimes.shape[1] < 3:
            pass
        elif vecEventTimes.shape[0] < 3:
            vecEventTimes = vecEventTimes.T
        else:
            raise Exception(
                "Input error: vecEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        vecEventTimes = np.reshape(vecEventTimes, (-1, 1))
    # define event starts
    vecEventStarts = vecEventTimes[:, 0]

    # check if number of events and spikes is sufficient
    if vecSpikeTimes.size < 3 or vecEventStarts.size < 3:
        if vecSpikeTimes.size < 3:
            strMsg1 = f"Number of spikes ({vecSpikeTimes.size}) is too few to calculate zeta; "
        else:
            strMsg1 = ""
        if vecEventStarts.size < 3:
            strMsg2 = f"Number of events ({vecEventStarts.size}) is too few to calculate zeta; "
        else:
            strMsg2 = ""
        logging.warning("zetatest: " + strMsg1 + strMsg2 + "defaulting to p=1.0")

        return vecTime, vecRate, dIFR

    # trial dur
    if dblUseMaxDur is None:
        dblUseMaxDur = np.min(np.diff(vecEventStarts))
    else:
        dblUseMaxDur = np.float64(dblUseMaxDur)
        assert dblUseMaxDur.size == 1 and dblUseMaxDur > 0, "dblUseMaxDur is not a positive scalar float"

    # parallel processing: to do
    boolUseParallel = False

    # %% get difference from uniform
    vecThisDeviation, vecThisSpikeFracs, vecThisFracLinear, vecThisSpikeTimes = getTempOffsetOne(
        vecSpikeTimes, vecEventStarts, dblUseMaxDur)
    intSpikeNum = vecThisSpikeTimes.size

    # check if sufficient spikes are present
    if vecThisDeviation.size < 3:
        logging.warning("ifr: too few spikes, returning empty variables")

        return vecTime, vecRate, dIFR

    # %% get multi-scale derivative
    intMaxRep = vecEventStarts.size
    dblMeanRate = (intSpikeNum/(dblUseMaxDur*intMaxRep))
    [vecRate, dMSD] = getMultiScaleDeriv(vecThisSpikeTimes, vecThisDeviation,
                                         dblSmoothSd=dblSmoothSd, dblMinScale=dblMinScale, dblBase=dblBase, dblMeanRate=dblMeanRate, dblUseMaxDur=dblUseMaxDur, boolParallel=boolParallel)

    # %% build output
    vecTime = dMSD['vecT']
    vecRate = vecRate  # unnecessary, but for clarity of output building
    dIFR = dict()
    dIFR['vecTime'] = vecTime
    dIFR['vecRate'] = vecRate
    dIFR['vecDeviation'] = vecThisDeviation
    dIFR['vecScale'] = dMSD['vecScale']
    return vecTime, vecRate, dIFR



import numpy as np
import time
import logging
import math
import matplotlib.pyplot as plt
import tkinter as tk
from zetapy.legacy import msd
from scipy import stats
from zetapy.legacy.dependencies import (flatten, getTempOffset, getGumbel, getPeak, getOnset,
                                 calculatePeths)


def getZeta(arrSpikeTimes, arrEventTimes, dblUseMaxDur=None, intResampNum=100, intPlot=0,
            intLatencyPeaks=2, tplRestrictRange=(-np.inf, np.inf),
            boolReturnRate=False, boolReturnZETA=False, boolVerbose=False, boolSuppressUpgradeMessage=False):
    """
    Calculates neuronal responsiveness index ZETA.

    Montijn, J.S., Seignette, K., Howlett, M.H., Cazemier, J.L., Kamermans, M., Levelt, C.N.,
    and Heimel, J.A. (2021). A parameter-free statistical test for neuronal responsiveness.
    eLife 10, e71969.

    Parameters
    ----------
    arrSpikeTimes : 1D array
        spike times (in seconds).
    arrEventTimes : 1D or 2D array
        event on times (s), or [T x 2] including event off times to calculate mean-rate difference.
    dblUseMaxDur : float
        window length for calculating ZETA: ignore all spikes beyond this duration after event onset
        (default: median of event onset to event onset)
    intResampNum : integer
        number of resamplings (default: 100)
    intPlot : int
        plotting switch (0: no plot, 1: plot figure) (default: 0)
    intLatencyPeaks : integer
        maximum number of latency peaks to return (1-4) (default: 2)
    tplRestrictRange : 2 element tuple
        temporal range within which to restrict onset/peak latencies (default: [-inf inf])
    boolReturnRate : boolean
        switch to return dictionary with spiking rate features
    boolReturnZETA : boolean
        switch to return dictionary with additional ZETA parameters
    boolVerbose : boolean
        switch to print progress messages (default: false)
    boolSuppressUpgradeMessage :  boolean
        switch to suppress the message to upgrade to zetatest (default: False)

    Returns
    -------
    dblZetaP : float
        p-value based on Zenith of Event-based Time-locked Anomalies
    arrLatencies : 1D array
        different latency estimates, number determined by intLatencyPeaks.
        If no peaks are detected, it returns NaNs
            1) Latency of ZETA
            2) Latency of largest z-score with inverse sign to ZETA
            3) Peak time of instantaneous firing rate
            4) Onset time of above peak, defined as the first crossing of peak half-height
    dZETA : dict (optional)
        additional parameters of ZETA test, return when using boolReturnZETA
            dblZETA; FDR-corrected responsiveness z-score (i.e., >2 is significant)
            dblD; temporal deviation value underlying ZETA
            dblP; p-value corresponding to ZETA
            dblPeakT; time corresponding to ZETA
            intPeakIdx; entry corresponding to ZETA
            dblMeanD; Cohen's D based on mean-rate stim/base difference
            dblMeanP; p-value based on mean-rate stim/base difference
            vecSpikeT: timestamps of spike times (corresponding to vecD)
            vecRealFrac; cumulative distribution of spike times
            vecRealFracLinear; linear baseline of cumulative distribution
            vecD; temporal deviation vector of data
            vecNoNormD; temporal deviation which is not mean subtracted
            matRandD; baseline temporal deviation matrix of jittered data
            dblD_InvSign; largest peak of inverse sign to ZETA (i.e., -ZETA)
            dblPeakT_InvSign; time corresponding to -ZETA
            intPeakIdx_InvSign; entry corresponding to -ZETA
            dblUseMaxDur; window length used to calculate ZETA
    dRate : dict (optional)
        additional parameters of the firing rate, return with boolReturnRate
            vecRate; instantaneous spiking rates (like a PSTH)
            vecT; time-points corresponding to vecRate (same as dZETA.vecSpikeT)
            vecM; Mean of multi-scale derivatives
            vecScale; timescales used to calculate derivatives
            matMSD; multi-scale derivatives matrix
            vecV; values on which vecRate is calculated (same as dZETA.vecZ)
            Data on the peak:
            dblPeakTime; time of peak (in seconds)
            dblPeakWidth; duration of peak (in seconds)
            vecPeakStartStop; start and stop time of peak (in seconds)
            intPeakLoc; spike index of peak (corresponding to dZETA.vecSpikeT)
            vecPeakStartStopIdx; spike indices of peak start/stop (corresponding to dZETA.vecSpikeT)
            Additionally, it will return peak onset latency (first crossing of peak half-height)
            dblOnset: latency for peak onset

    Original code by Jorrit Montijn, ported to python by Alexander Heimel & Guido Meijer

    Version history:
    2.5 - 17 June 2020 Jorrit Montijn, translated to python by Alexander Heimel
    2.5.1 - 18 February 2022 Bugfix by Guido Meijer of 1D arrEventTimes
    2.6 - 20 February 2022 Refactoring of python code by Guido Meijer
    """
    
    # boolSuppressUpgradeMessage=False
    if boolSuppressUpgradeMessage is None:
        boolSuppressUpgradeMessage=False
    
    if boolSuppressUpgradeMessage is False:
        print('Please upgrade to the new version of the zeta-test called zetatest, as it has improved statistical sensitivity.\nIf you wish to continue using getZeta for legacy purposes and want to suppress this message, you can call getZeta with the optional parameter boolSuppressUpgradeMessage=True')
        
    # ensure arrEventTimes is a N x 2 array
    if len(arrEventTimes.shape) > 1:
        boolStopSupplied = True
        if np.shape(arrEventTimes)[1] > 2:
            arrEventTimes = np.transpose(arrEventTimes)
    else:
        boolStopSupplied = False
        arrEventTimes = np.vstack((arrEventTimes, np.zeros(arrEventTimes.shape))).T

    # trial dur
    if dblUseMaxDur is None:
        dblUseMaxDur = np.median(np.diff(arrEventTimes[:, 0]))

    # build onset/offset vectors
    arrEventStarts = arrEventTimes[:, 0]

    # prepare interpolation points
    intMaxRep = np.shape(arrEventTimes)[0]
    cellSpikeTimesPerTrial = [None] * intMaxRep

    # go through trials to build spike time vector
    for intEvent in range(intMaxRep):
        # get times
        dblStartT = arrEventStarts[intEvent]
        dblStopT = dblStartT + dblUseMaxDur

        # build trial assignment
        cellSpikeTimesPerTrial[intEvent] = arrSpikeTimes[(arrSpikeTimes < dblStopT)
                                                         & (arrSpikeTimes > dblStartT)] - dblStartT

    # get spikes in fold
    vecSpikeT = np.array(sorted(flatten([0, cellSpikeTimesPerTrial, dblUseMaxDur])))
    intSpikes = vecSpikeT.shape[0]

    # run normal
    vecOrigDiff, vecRealFrac, vecRealFracLinear = getTempOffset(vecSpikeT, arrSpikeTimes,
                                                                arrEventStarts, dblUseMaxDur)

    # mean subtract difference
    vecRealDiff = vecOrigDiff - np.mean(vecOrigDiff)

    # run bootstraps
    hTic = time.time()
    matRandDiff = np.empty((intSpikes, intResampNum))
    matRandDiff[:] = np.nan
    for intResampling in range(intResampNum):
        # msg
        if boolVerbose and ((time.time()-hTic) > 5):
            print('Now at resampling %d/%d' % (intResampling, intResampNum))
            hTic = time.time()

        # get random subsample
        vecStimUseOnTime = (arrEventStarts + 2 * dblUseMaxDur
                            * ((np.random.rand(arrEventStarts.shape[0]) - 0.5) * 2))

        # get temp offset
        vecRandDiff, vecRandFrac, vecRandFracLinear = getTempOffset(vecSpikeT, arrSpikeTimes,
                                                                    vecStimUseOnTime, dblUseMaxDur)

        # assign data
        matRandDiff[:, intResampling] = vecRandDiff - np.mean(vecRandDiff)

    # calculate measure of effect size (for equal n, d' equals Cohen's d)
    if (len(vecRealDiff) < 3) | (arrSpikeTimes.shape[0] < 10):
        if boolVerbose:
            logging.warning('Insufficient samples to calculate zeta')
        dblZetaP = 1
        arrLatencies = np.array([np.nan] * intLatencyPeaks)
        dZETA = dict()
        dRate = dict()
        if (boolReturnZETA & boolReturnRate):
            return dblZetaP, arrLatencies, dZETA, dRate
        elif boolReturnZETA:
            return dblZetaP, arrLatencies, dZETA
        elif boolReturnRate:
            return dblZetaP, arrLatencies, dRate
        else:
            return dblZetaP, arrLatencies

    # find highest peak and retrieve value
    vecMaxRandD = np.max(np.abs(matRandDiff), 0)
    dblRandMu = np.mean(vecMaxRandD)
    dblRandVar = np.var(vecMaxRandD, ddof=1)
    intZETALoc = np.argmax(np.abs(vecRealDiff))
    dblPosD = np.max(np.abs(vecRealDiff))  # Can be combined with line above

    # get location
    dblMaxDTime = vecSpikeT[intZETALoc]
    dblD = vecRealDiff[intZETALoc]

    # calculate statistical significance using Gumbel distribution
    if boolVerbose:
        print('Python: Gumbel %0.7f, %0.7f, %0.7f' % (dblRandMu, dblRandVar, dblPosD))
    dblZetaP, dblZETA = getGumbel(dblRandMu, dblRandVar, dblPosD)

    # find peak of inverse sign
    intPeakLocInvSign = np.argmax(-np.sign(dblD)*vecRealDiff)
    dblMaxDTimeInvSign = vecSpikeT[intPeakLocInvSign]
    dblD_InvSign = vecRealDiff[intPeakLocInvSign]

    if boolStopSupplied:
        # calculate mean-rate difference
        vecEventStops = arrEventTimes[:, 1]
        vecStimHz = np.zeros(intMaxRep)
        vecBaseHz = np.zeros(intMaxRep)
        dblMedianBaseDur = np.median(arrEventStarts[1:] - vecEventStops[0:-1])

        # go through trials to build spike time vector
        for intEvent in range(intMaxRep):
            # get times
            dblStartT = arrEventStarts[intEvent]
            dblStopT = dblStartT + dblUseMaxDur
            dblPreT = dblStartT - dblMedianBaseDur

            # build trial assignment
            vecStimHz[intEvent] = (np.sum((arrSpikeTimes < dblStopT) & (arrSpikeTimes > dblStartT))
                                   / (dblStopT - dblStartT))
            vecBaseHz[intEvent] = (np.sum((arrSpikeTimes < dblStartT) & (arrSpikeTimes > dblPreT))
                                   / dblMedianBaseDur)

        # get metrics
        dblMeanD = np.mean(vecStimHz - vecBaseHz) / ((np.std(vecStimHz) + np.std(vecBaseHz)) / 2)
        dblMeanP = stats.ttest_rel(vecStimHz, vecBaseHz)[1]

    if intPlot == 1:
        # logging.warning('Plotting is not translated to python yet')

        # Plot maximally 50 traces
        intPlotIters = np.min([arrEventStarts.shape[0], 50])

        # Calculate optimal DPI depending on the monitor size
        screen_width = tk.Tk().winfo_screenwidth()
        dpi = screen_width / 15

        # Create figure
        f, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = plt.subplots(2, 3, figsize=(12, 6), dpi=dpi)

        # Plot spike raster
        for i, t in enumerate(arrEventStarts[:intPlotIters]):
            idx = np.bitwise_and(arrSpikeTimes >= t, arrSpikeTimes <= t + dblUseMaxDur)
            event_spks = arrSpikeTimes[idx]
            ax1.vlines(event_spks - t, i + 1, i, color='k', lw=0.3)
        ax1.set(xlabel='Time from event (s)', ylabel='Trial #', title='Spike raster plot')

        # Get peri-event time histogram
        peth, binned_spikes = calculatePeths(arrSpikeTimes, np.ones(arrSpikeTimes.shape), [1],
                                             arrEventStarts, pre_time=0, post_time=dblUseMaxDur,
                                             bin_size=dblUseMaxDur/25, smoothing=0)
        ax2.errorbar(peth['tscale'], peth['means'][0, :], yerr=peth['sems'])
        ax2.set(xlabel='Time from event (s)', ylabel='spks/s',
                title='Mean spiking over trials')

        ax3.plot(vecSpikeT, vecRealFrac)
        ax3.plot(vecSpikeT, vecRealFracLinear, color=[0.7, 0.7, 0.7])
        ax3.set(xlabel='Time from event (s)', ylabel='Fractional position \n of spike in trial',
                title='Real data')

        for i in range(intPlotIters):
            ax4.plot(vecSpikeT, matRandDiff[:, i], color=[0.7, 0.7, 0.7])
        ax4.plot(vecSpikeT, vecRealDiff)
        ax4.plot(dblMaxDTime, vecRealDiff[intZETALoc], 'bx')
        ax4.plot(dblMaxDTimeInvSign, vecRealDiff[intPeakLocInvSign], 'b*')
        ax4.set(xlabel='Time from event (s)', ylabel='Offset of data from linear (s)')
        if boolStopSupplied:
            ax4.set(title=f'Legacy ZETA={dblZETA:.3f} (p={dblZetaP:.3f}), d(Hz)={dblMeanD:.3f} (p={dblMeanP:.3f})')
        else:
            ax4.set(title=f'Legacy ZETA={dblZETA:.3f} (p={dblZetaP:.3f})')

    else:
        ax5, ax6 = [], []

    # calculate MSD if significant
    if intLatencyPeaks > 0:
        # get average of multi-scale derivatives, and rescaled to instantaneous spiking rate
        dblMeanRate = intSpikes / (dblUseMaxDur * intMaxRep)
        vecRate, dRate = msd.getMultiScaleDeriv(vecSpikeT, vecRealDiff, intPlot=intPlot,
                                                dblMeanRate=dblMeanRate, dblUseMaxDur=dblUseMaxDur,
                                                axs=[ax5, ax6])
    else:
        dRate = None

    # calculate MSD statistics
    if dRate is not None and intLatencyPeaks > 0:
        # get MSD peak
        (dblPeakRate, dblPeakTime, dblPeakWidth, vecPeakStartStop,
         intPeakLoc, vecPeakStartStopIdx) = getPeak(vecRate, vecSpikeT, tplRestrictRange)

        dRate['dblPeakRate'] = dblPeakRate
        dRate['dblPeakTime'] = dblPeakTime
        dRate['dblPeakWidth'] = dblPeakWidth
        dRate['vecPeakStartStop'] = vecPeakStartStop
        dRate['intPeakLoc'] = intPeakLoc
        dRate['vecPeakStartStopIdx'] = vecPeakStartStopIdx

        if not math.isnan(dblPeakTime):
            # assign array data
            if intLatencyPeaks > 3:
                # get onset
                dblOnset, dblOnsetVal, dblBaseVal, dblPeakT = getOnset(
                    vecRate, vecSpikeT, dblPeakTime, tplRestrictRange)
                dRate['dblOnset'] = dblOnset
                arrLatencies = np.array([dblMaxDTime, dblMaxDTimeInvSign, dblPeakTime, dblOnset])
                vecLatencyVals = np.array([vecRate[intZETALoc], vecRate[intPeakLocInvSign],
                                           vecRate[intPeakLoc], dblOnsetVal], dtype=object)
            else:
                dRate['dblOnset'] = np.nan
                arrLatencies = np.array([dblMaxDTime, dblMaxDTimeInvSign, dblPeakTime])
                vecLatencyVals = np.array([vecRate[intZETALoc], vecRate[intPeakLocInvSign],
                                           vecRate[intPeakLoc]], dtype=object)
            arrLatencies = arrLatencies[0:intLatencyPeaks]
            vecLatencyVals = vecLatencyVals[0:intLatencyPeaks]
            if intPlot == 1:

                ax6.plot(dblPeakTime, vecRate[intPeakLoc], 'gx')
                ax6.plot(dblMaxDTime, vecRate[intZETALoc], 'bx')
                ax6.plot(dblMaxDTimeInvSign, vecRate[intPeakLocInvSign], 'b*')
                if intLatencyPeaks > 3:
                    ax6.plot(dblOnset, dblOnsetVal, 'rx')
                    ax6.set(title=f'ZETA={dblMaxDTime*1000:.0f}ms,-ZETA={dblMaxDTimeInvSign*1000:.0f}ms,'
                            'Pk={dblPeakTime*1000:.0f}ms,On={dblOnset*1000:.2f}ms')
                else:
                    ax6.set(title=f'ZETA={dblMaxDTime*1000:.0f}ms,-ZETA={dblMaxDTimeInvSign*1000:.0f}ms')

                """
     
                if intPlot > 3
                    vecHandles = get(gcf,'children');
                    ptrFirstSubplot = vecHandles(find(contains(get(vecHandles,'type'),'axes'),1,'last'));
                    axes(ptrFirstSubplot);
                    vecY = get(gca,'ylim');
                    hold on;
                    if intLatencyPeaks > 3,plot(dblOnset*[1 1],vecY,'r--');end
                    plot(dblPeakTime*[1 1],vecY,'g--');
                    plot(dblMaxDTime*[1 1],vecY,'b--');
                    plot(dblMaxDTimeInvSign*[1 1],vecY,'b-.');
                    hold off
                end
                """
        else:
            # placeholder peak data
            dRate['dblOnset'] = np.nan
            arrLatencies = np.array([np.nan] * intLatencyPeaks)
            vecLatencyVals = np.array([np.nan] * intLatencyPeaks)
    else:
        arrLatencies = []
        vecLatencyVals = []

    if intPlot == 1:
        plt.tight_layout()

    # build optional output dictionary
    dZETA = dict()
    dZETA['dblZeta'] = dblZETA
    dZETA['dblD'] = dblD
    dZETA['dblP'] = dblZetaP
    dZETA['dblPeakT'] = dblMaxDTime
    dZETA['intPeakIdx'] = intZETALoc
    if boolStopSupplied:
        dZETA['dblMeanD'] = dblMeanD
        dZETA['dblMeanP'] = dblMeanP
    dZETA['vecSpikeT'] = vecSpikeT
    dZETA['vecRealFrac'] = vecRealFrac
    dZETA['vecRealFracLinear'] = vecRealFracLinear
    dZETA['vecD'] = vecRealDiff
    dZETA['vecNoNormD'] = vecOrigDiff
    dZETA['matRandD'] = matRandDiff
    dZETA['dblD_InvSign'] = dblD_InvSign
    dZETA['dblPeakT_InvSign'] = dblMaxDTimeInvSign
    dZETA['intPeakIdx_InvSign'] = intPeakLocInvSign
    dZETA['dblUseMaxDur'] = dblUseMaxDur

    if (boolReturnZETA & boolReturnRate):
        return dblZetaP, arrLatencies, dZETA, dRate
    elif boolReturnZETA:
        return dblZetaP, arrLatencies, dZETA
    elif boolReturnRate:
        return dblZetaP, arrLatencies, dRate
    else:
        return dblZetaP, arrLatencies


def getIFR(arrSpikeTimes, arrEventStarts, dblUseMaxDur=None, intSmoothSd=5, dblMinScale=None,
           dblBase=1.5, intPlot=0, boolVerbose=True):
    """Returns multi-scale derivative. Syntax:
       [vecMSD,sMSSD] = getMultiScaleSpikeDeriv(arrSpikeTimes,arrEventStarts,dblUseMaxDur,intSmoothSd,dblMinScale,dblBase,intPlot,boolVerbose)
    Required input:
        - arrSpikeTimes [S x 1]: spike times (s)
        - arrEventStarts [T x 1]: event on times (s), or [T x 2] including event off times
        - dblUseMaxDur: float (s), ignore all spikes beyond this duration after stimulus onset
                                    [default: median of trial start to trial start]

    Optional inputs:
        - intSmoothSd: Gaussian SD of smoothing kernel (in # of bins) [default: 3]
        - dblMinScale: minimum derivative scale in seconds [default: 1/1000]
        - dblBase: critical value for locally dynamic derivative [default: 4]
        - intPlot: integer, plotting switch (0=none, 1=plot)
        - boolVerbose: boolean, switch to print messages

    Outputs:
        - vecMSprime; Multi-scale derivative
        - sMSSD; structure with fields:
            - vecMSD;
            - vecSpikeT;
            - vecFracs;
            - vecLinear;
            - vecDiff;
            - vecScale;
            - matSmoothMSprime;
            - matMSprime;

    Version history:
    1.0 - June 24, 2020 Created by Jorrit Montijn, translated to python by Alexander Heimel
    """

    if dblMinScale == None:
        dblMinScale = np.round(np.log(1/1000) / np.log(dblBase))

    if dblUseMaxDur == None:
        dblUseMaxDur = np.median(np.diff(arrEventStarts[:, 0]))

    # prepare normalized spike times
    # pre-allocate
    intMaxRep = np.shape(arrEventStarts)[0]
    cellSpikeTimesPerTrial = [None] * intMaxRep

    # go through trials to build spike time vector
    for intEvent in range(intMaxRep):
        # get times
        dblStartT = arrEventStarts[intEvent, 0]
        dblStopT = dblStartT + dblUseMaxDur

        # build trial assignment
        cellSpikeTimesPerTrial[intEvent] = arrSpikeTimes[(arrSpikeTimes < dblStopT)
                                                         & (arrSpikeTimes > dblStartT)] - dblStartT

    # get spikes in fold
    vecSpikeT = np.array(sorted(flatten(cellSpikeTimesPerTrial)))

    # get difference from uniform
    vecFracs = np.linspace(0, 1, vecSpikeT.shape[0])
    vecLinear = vecSpikeT / np.max(vecSpikeT)
    vecDiff = vecFracs - vecLinear
    vecDiff = vecDiff - np.mean(vecDiff)

    # get multi-scale derivative
    vecMSD, sMSD = msd.getMultiScaleDeriv(vecSpikeT, vecDiff, intSmoothSd, dblMinScale, dblBase, intPlot)

    sMSD['vecSpikeT'] = vecSpikeT
    sMSD['vecFracs'] = vecFracs
    sMSD['vecLinear'] = vecLinear
    sMSD['vecDiff'] = vecDiff

    return vecMSD, sMSD



"""Run example ZETA-test.

This code loads data from an example cell and performs some analyses as a tutorial/

Please note that the MATLAB version is usually faster than zetapy due to the amount of
vectorized operations in the zeta-test that are computed very efficiently in MATLAB,
but less so in numpy. If anyone has any suggestions to improve compution times in python,
please contact me by e-mail or on the github repository: https://github.com/JorritMontijn/zetapy

Don't hesitate to leave any questions, bug reports or comments there either!

Version history:
1.0 - 17 June 2020 Translated to python by Alexander Heimel
1.1 - 23 Feb 2021 Refactoring of python code by Guido Meijer
2.0 - 29 Aug 2023 New translation to python by Jorrit Montijn
3.0 - 30 Oct 2023 Ported new two-sample tests to Python [by JM]

"""

import scipy.io
from zetapy import ifr, zetatest, zetatstest, zetatest2, zetatstest2, plotzeta, plottszeta, plotzeta2, plottszeta2
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
from scipy.signal import convolve
import time

# %% note on performance
'''
The code uses vectorization as much as possible, but since the ZETA-test is based on bootstraps, it
might still take a couple of seconds to compute. Parallel processing in python is not yet implemented,
so if an expert in parallel computing in python reads this who is willing to contribute: please contact me!

In the examples below, I recorded the following processing times on my PC with default parameters:

MATLAB with parallel processing:
    zeta-test: 0.56 s
    time-series zeta-test: 2.77 s

MATLAB without parallel processing:
    zeta-test: 1.09 s
    time-series zeta-test: 19.88 s

Python without parallel processing:
    zeta-test: 1.41 s
    time-series zeta-test: 32.24 s
'''

# %% load and prepare some example data
# load data for example cell
strDataFile = os.path.join(os.path.dirname(__file__), 'ExampleDataZetaTest.mat')
dLoad = scipy.io.loadmat(strDataFile)

# retrieve the spike times as an array from the field in dNeuron
vecSpikeTimes1 = dLoad['sNeuron']['SpikeTimes'][0][0]
vecSpikeTimes2 = dLoad['sNeuron']['SpikeTimes'][0][1]

# load stimulation information
sStim = dLoad['sStim']
vecStimulusStartTimes = sStim['StimOnTime'][0][0][0]  # unpacking Matlab array
vecStimulusStopTimes = sStim['StimOffTime'][0][0][0]  # unpacking Matlab array
vecOrientation = sStim['Orientation'][0][0][0]  # unpacking Matlab array

# %% calculate instantaneous firing rate without performing the ZETA-test
# if we simply want to plot the neuron's response, we can use:
vecTime, vecRate, dIFR = ifr(vecSpikeTimes1, vecStimulusStartTimes)

# plot results
f, ax = plt.subplots(1, figsize=(6, 4))
ax.plot(vecTime, vecRate)
ax.set(xlabel='Time after event (s)', ylabel='Instantaneous spiking rate (Hz)')
ax.set(title="A simple plot of the neuron's rate using ifr()")

# %% run the ZETA-test with default parameters
# set random seed
np.random.seed(1)

# run test
t = time.time()
dblZetaP = zetatest(vecSpikeTimes1, vecStimulusStartTimes)[0]  # use [0] to return only the p-value
dblElapsedT = time.time() - t

print(f'\nDefault parameters (elapsed time: {dblElapsedT:.2f} s):\np-value: {dblZetaP}')

# %% run the ZETA-test with specified parameters
# set random seed
np.random.seed(1)

# use minimum of trial-to-trial durations as analysis window size
dblUseMaxDur = np.min(np.diff(vecStimulusStartTimes))

# 50 random resamplings should give us a good enough idea if this cell is responsive.
# If the p-value is close to 0.05, we should increase this number.
intResampNum = 50

# what size of jittering do we want? (multiple of dblUseMaxDur; default is 2.0)
dblJitterSize = 2.0

# Do we want to plot the results?
boolPlot = True

# do we want to restrict the peak detection to for example the time during stimulus?
# Then put (0 1) here.
tplRestrictRange = (0, np.inf)

# do we want to compute the instantaneous firing rate?
boolReturnRate = True

# create a T by 2 array with stimulus onsets and offsets so we can also compute the t-test
arrEventTimes = np.transpose(np.array([vecStimulusStartTimes, vecStimulusStopTimes]))

# then run ZETA with those parameters
t = time.time()
dblZetaP, dZETA, dRate = zetatest(vecSpikeTimes1, arrEventTimes,
                                                dblUseMaxDur=dblUseMaxDur,
                                                intResampNum=intResampNum,
                                                dblJitterSize=dblJitterSize,
                                                boolPlot=boolPlot,
                                                tplRestrictRange=tplRestrictRange,
                                                boolReturnRate=boolReturnRate)

dblElapsedT2 = time.time() - t
print(f"\nSpecified parameters (elapsed time: {dblElapsedT2:.2f} s): \
      \nzeta-test p-value: {dblZetaP}\nt-test p-value:{dZETA['dblMeanP']}")


# Note on the latencies: while the peaks of ZETA and -ZETA can be useful for diagnostic purposes,
# they are difficult to interpret, so we suggest sticking to the peak time (vecLatencies[2]),
# which is more easily interpretable. Please read the paper for more information.
vecLatencies = dZETA['vecLatencies']

# %% run the time-series zeta-test
# take subselection of data
intUseTrialNum = 480
vecStimulusStartTimesTs = vecStimulusStartTimes[0:intUseTrialNum]
vecStimulusStopTimesTs = vecStimulusStopTimes[0:intUseTrialNum]
vecOrientationTs = vecOrientation[0:intUseTrialNum]
arrEventTimesTs = np.transpose(np.array([vecStimulusStartTimesTs, vecStimulusStopTimesTs]))

# first transform the data to time-series
print('\nRunning time-series zeta-test; This will take around 40 seconds\n')
dblStartT = 0
dblEndT = vecStimulusStopTimesTs[-1] + dblUseMaxDur*5
dblSamplingRate = 50.0  # simulate acquisition rate
dblSampleDur = 1/dblSamplingRate
vecTimestamps = np.arange(dblStartT, dblEndT+dblSampleDur, dblSampleDur)
vecSpikesBinned = np.histogram(vecSpikeTimes1, bins=vecTimestamps)[0]
vecTimestamps = vecTimestamps[0:-1]
dblSmoothSd = 1.0
intSmoothRange = 2*np.ceil(dblSmoothSd).astype(int)
vecFilt = norm.pdf(range(-intSmoothRange, intSmoothRange+1), 0, dblSmoothSd)
vecFilt = vecFilt / sum(vecFilt)

# pad array
intPadSize = np.floor(len(vecFilt)/2).astype(int)
vecData1 = np.pad(vecSpikesBinned, ((intPadSize, intPadSize)), 'edge')

# filter
vecData1 = convolve(vecData1, vecFilt, 'valid')

# set random seed
np.random.seed(1)

# time-series zeta-test with default parameters
t = time.time()
dblTsZetaP = zetatstest(vecTimestamps, vecData1, vecStimulusStartTimesTs)[0]
dblElapsedT3 = time.time() - t
print(f"\nDefault parameters (elapsed time: {dblElapsedT3:.2f} s):\ntime-series zeta-test p-value: {dblTsZetaP}\n")

# %% run time-series zeta-test with specified parameters
# set random seed
np.random.seed(1)
t = time.time()

# run test
print('\nRunning time-series zeta-test with specified parameters; This will take around 40 seconds\n')
dblTsZetaP2, dZetaTs = zetatstest(vecTimestamps, vecData1, arrEventTimesTs,
                                  dblUseMaxDur=None, intResampNum=100, boolPlot=True,
                                  dblJitterSize=2.0, boolDirectQuantile=False)

dblElapsedT4 = time.time() - t
print(f"\nSpecified parameters (elapsed time: {dblElapsedT4:.2f} s): \
      \ntime-series zeta-test p-value: {dblTsZetaP2}\nt-test p-value:{dZetaTs['dblMeanP']}")


# %% run the two-sample ZETA-test
#case 1: are neurons 1 & 2 responding differently to a set of visual stimuli?
np.random.seed(1)
print('\nRunning two-sample zeta-test on two neurons, same stimuli\n')
t = time.time()
intTrials = 240; #that's already more than enough
intResampNum = 500; #the two-sample test is more variable, as it depends on differences, so it requires more resamplings
dblZetaTwoSample,dZETA2 = zetatest2(vecSpikeTimes1,arrEventTimes[0:intTrials,:],vecSpikeTimes2,arrEventTimes[0:intTrials,:],
                                    dblUseMaxDur,intResampNum,boolPlot=True)
dblElapsedT5 = time.time() - t
print(f"\nAre two neurons responding differently? (elapsed time: {dblElapsedT5:.2f} s): \
      \ntwo-sample zeta-test p-value: {dblZetaTwoSample}\nt-test p-value:{dZETA2['dblMeanP']}")


# case 2a: is neuron 1 responding differently to gratings oriented at 0 and 90 degrees?
vecTrials1 = vecOrientation==0;
vecTrials2 = vecOrientation==90;
print('\nRunning two-sample zeta-test on one neuron, different stimuli\n')
t = time.time()
dblZetaTwoSample2a,dZETA2a = zetatest2(vecSpikeTimes1,arrEventTimes[vecTrials1,:],vecSpikeTimes1,arrEventTimes[vecTrials2,:],
                                       dblUseMaxDur,intResampNum,boolPlot=True)
dblElapsedT6 = time.time() - t
print(f"\nIs neuron 1 responding differently to 0 and 90 degree stimuli? (elapsed time: {dblElapsedT6:.2f} s): \
      \ntwo-sample zeta-test p-value: {dblZetaTwoSample2a}\nt-test p-value:{dZETA2a['dblMeanP']}")

#case 2b: is neuron 2 responding differently to gratings oriented at 0 and 90 degrees?
t = time.time()
intResampNum = 1000; #this difference is close to our threshold p=0.05, so we're increasing the number of bootstraps
[dblZetaTwoSample2b,dZETA2b] = zetatest2(vecSpikeTimes2,arrEventTimes[vecTrials1,:],vecSpikeTimes2,arrEventTimes[vecTrials2,:],
                                         dblUseMaxDur,intResampNum,boolPlot=True)
dblElapsedT7 = time.time() - t
print(f"\nIs neuron 2 responding differently to 0 and 90 degree stimuli? (elapsed time: {dblElapsedT7:.2f} s): \
      \ntwo-sample zeta-test p-value: {dblZetaTwoSample2b}\nt-test p-value:{dZETA2b['dblMeanP']}")


# %% finally, the two-sample time-series ZETA test
#get trials
vecTrials1 = vecOrientationTs==0
vecTrials2 = vecOrientationTs==90

#get data for neuron 2
dblStartT = 0
dblEndT = vecStimulusStopTimesTs[-1] + dblUseMaxDur*5
dblSamplingRate = 50.0  # simulate acquisition rate
dblSampleDur = 1/dblSamplingRate
vecTimestamps = np.arange(dblStartT, dblEndT+dblSampleDur, dblSampleDur)
vecSpikesBinned = np.histogram(vecSpikeTimes2, bins=vecTimestamps)[0]
vecTimestamps = vecTimestamps[0:-1]
dblSmoothSd = 1.0
intSmoothRange = 2*np.ceil(dblSmoothSd).astype(int)
vecFilt = norm.pdf(range(-intSmoothRange, intSmoothRange+1), 0, dblSmoothSd)
vecFilt = vecFilt / sum(vecFilt)

# pad array
intPadSize = np.floor(len(vecFilt)/2).astype(int)
vecData2 = np.pad(vecSpikesBinned, ((intPadSize, intPadSize)), 'edge')

# filter
vecData2 = convolve(vecData2, vecFilt, 'valid')

#set parameters
intResampNum=250
boolPlot=True
boolDirectQuantile=False
dblSuperResFactor=100
#case 1: are neurons 1 & 2 responding differently to a set of visual stimuli?
np.random.seed(1)
print('\nRunning two-sample time-series zeta-test on two neurons, same stimuli\n')
t = time.time()
dblTsZetaTwoSample,dTSZETA2 = zetatstest2(vecTimestamps, vecData1, arrEventTimesTs, vecTimestamps, vecData2, arrEventTimesTs,
                dblUseMaxDur, intResampNum, boolPlot, boolDirectQuantile, dblSuperResFactor)

dblElapsedT8 = time.time() - t
print(f"\nAre two neurons responding differently? (elapsed time: {dblElapsedT8:.2f} s): \
      \ntwo-sample time-series zeta-test p-value: {dblTsZetaTwoSample}\nt-test p-value:{dTSZETA2['dblMeanP']}")


#case 2: is neuron 1 responding differently to gratings oriented at 0 and 90 degrees?
print('\nRunning two-sample time-series zeta-test on one neuron, different stimuli\n')
t = time.time()
dblTsZetaTwoSampleB,sTSZETA2B = zetatstest2(vecTimestamps,vecData1,arrEventTimesTs[vecTrials1,:],vecTimestamps,vecData1,arrEventTimesTs[vecTrials2,:],
                dblUseMaxDur, intResampNum, boolPlot, boolDirectQuantile, dblSuperResFactor)
dblElapsedT9 =  time.time() - t
print(f"\nIs neuron 1 responding differently to 0 and 90 degree stimuli? (elapsed time: {dblElapsedT9:.2f} s): \
      \ntwo-sample time-series zeta-test p-value: {dblTsZetaTwoSampleB}\nt-test p-value:{sTSZETA2B['dblMeanP']}")
